/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.multirefac.gathering;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.IJavaSearchScope;
import org.eclipse.jdt.core.search.SearchEngine;
import org.eclipse.jdt.core.search.SearchParticipant;
import org.eclipse.jdt.core.search.SearchPattern;
import org.intoJ.inferType.exception.UserAbortedException;
import org.intoJ.inferType.model.collection.DESet_Key;
import org.intoJ.inferType.util.SearchResultCollector;
import org.intoJ.inferType.util.Util_ITypes;

/**
 * DeclarationElementFinder.
 * 
 * This class is used to find declarations of a certain type - or a number of
 * types - inside a given scope.
 * 
 * A declaration element is defined as 
 * -> variables declared with the given type (this includes fields, local variables, 
 *    parameters, variables inside for loops and in catch blocks) 
 * -> methods which have the given type as their return type.
 * 
 * Note that the classes themselves are not found, even if they are, so to
 * speak, declaration elements on which something may be invoked: new
 * ClassA().doSomething() or this.doSomething().
 * 
 * The first case, constructors, are similar to methods, but their return type
 * cannot be changed. The second case, "this", is similar to fields, but the
 * type cannot be changed either.
 * 
 * As we focus on the re-declaration of those types, it makes no sense to look
 * at either "this" or constructors.
 * 
 * @version: $Date: 2007/07/28 12:12:55 $Author: kegel $Revision: 1.1 $
 * @author Philip Mayer
 *  
 */
public class DeclarationElementFinder {

	public static DESet_Key findDeclarationSet(IType type, IJavaSearchScope scope, IProgressMonitor monitor, int work) throws UserAbortedException,
			CoreException {

		Set typeSet = new HashSet();
		typeSet.add(type);
		return findDeclarationSet(typeSet, scope, monitor, work);
	}

	public static DESet_Key findDeclarationSet(Set setOfITypes, IJavaSearchScope scope, IProgressMonitor monitor, int work) throws CoreException,
			UserAbortedException {

		if (setOfITypes.isEmpty())
			return new DESet_Key();

		Collection compilationUnits = getCompilationUnitsOfTypes(scope, setOfITypes);
		return findDeclarationSet(setOfITypes, compilationUnits, monitor, work, true);
	}

	public static DESet_Key findDeclarationSet(IType iType, Collection compilationUnits, IProgressMonitor monitor, int work, boolean startDEUpdater)
			throws UserAbortedException {

		Set set = new HashSet();
		set.add(iType);
		return findDeclarationSet(set, compilationUnits, monitor, work, startDEUpdater);
	}

	/**
	 * Determines all declaration elements declared with one of the given types and
	 * situated in one of the given compilation untis. 
	 * Declaration elements are variables (fields, parameters or local variables) 
	 * declared with a relevant type or methods whose return type is a relevant type.
	 * 
	 * @param setOfITypes the types
	 * @param compilationUnits collection of ICompilationUnits where the finder has to search for DEs
	 * @param monitor the monitor
	 * @param work the number of work units
	 * @param startDEUpdater should the DeclarationElementUpdater be startet after determining the DEs
	 * 
	 * @return a DeclarationSet
	 *  
	 * @throws UserAbortedException
	 * @throws JavaModelException
	 */
	public static DESet_Key findDeclarationSet(Set setOfITypes, Collection compilationUnits, IProgressMonitor monitor, int work, boolean startDEUpdater)
			throws UserAbortedException {

		DESet_Key declarationElements = new DESet_Key();

		// parse compilation units for variables (fields, parameters and local
		// variables) declared with a relevant type and methods whose return
		// type is of the relevant type.
		int stepWidth = compilationUnits.size() == 0 ? 0 : work / compilationUnits.size();
		Iterator icuIter = compilationUnits.iterator();
		while (icuIter.hasNext()) {
			Util_ITypes u = new Util_ITypes();
			CompilationUnit currentCU = u.parseCompilationUnit((ICompilationUnit) icuIter.next(), true);

			// Collecting parameter pattern: declaredElementsMap will be filled
			// by the visitor
			currentCU.accept(new DeclarationElementVisitor(setOfITypes, declarationElements));

			monitor.worked(stepWidth);

			if (monitor.isCanceled())
				throw new UserAbortedException("Operation canceled", "The operation was canceled.");
		}

		return declarationElements;
	}

	/**
	 * Determines the set of corresponding ICompilationUnits to a set of ITypes.
	 * 
	 * Please note that local (I)types are found as well.
	 * 
	 * @param scope the search scope
	 * @param setOfTypes set of ITypes
	 * 
	 * @return collection of ICompilationUnits
	 * 
	 * @throws CoreException
	 */
	public static Collection getCompilationUnitsOfTypes(IJavaSearchScope scope, Set setOfTypes) throws CoreException {

		// get compilation units with references to the relevant types
		// first create an OR-Search-Pattern
		Iterator typesIterator = setOfTypes.iterator();
		SearchPattern searchPattern = SearchPattern.createPattern((IType) typesIterator.next(), IJavaSearchConstants.ALL_OCCURRENCES);

		while (typesIterator.hasNext()) {
			IType currentType = (IType) typesIterator.next();
			SearchPattern tempPattern = SearchPattern.createPattern(currentType, IJavaSearchConstants.ALL_OCCURRENCES);
			searchPattern = SearchPattern.createOrPattern(tempPattern, searchPattern);
		}

		// then search
		SearchResultCollector compilationUnitCollector = new SearchResultCollector();
		SearchParticipant searchParticipant = SearchEngine.getDefaultSearchParticipant();
		SearchEngine searchEngine = new SearchEngine();

		searchEngine.search(searchPattern, new SearchParticipant[] { searchParticipant }, scope, compilationUnitCollector, null);

		Collection compilationUnits = compilationUnitCollector.getCompilationUnits();

		return compilationUnits;
	}
}