/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.actions;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IField;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.ILocalVariable;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.ISourceRange;
import org.eclipse.jdt.core.ISourceReference;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeRoot;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.ImportDeclaration;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.ParameterizedType;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SimpleType;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jdt.core.dom.WildcardType;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.jdt.internal.ui.javaeditor.CompilationUnitEditor;
import org.eclipse.jdt.internal.ui.javaeditor.JavaTextSelection;
import org.eclipse.jface.action.IAction;
import org.eclipse.jface.text.IDocument;
import org.eclipse.jface.text.ITextSelection;
import org.eclipse.jface.text.TextSelection;
import org.eclipse.swt.widgets.Event;
import org.eclipse.ui.IActionDelegate2;
import org.eclipse.ui.IEditorActionDelegate;
import org.eclipse.ui.IEditorPart;
import org.intoJ.inferType3.internal.constraintModel.util.Scope;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElementVisitor;
import org.intoJ.inferType3.internal.ui.ImportScopeDialog;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * 
 * This action can be executed by the user by selecting "Infer Type" from the context menus of the
 * editor.
 *
 * @author Andreas Mei&szlig;ner
 * @author Hannes Kegel
 */
public class InferTypeOnSelectionAction extends AbstractInferTypeAction implements IActionDelegate2, IEditorActionDelegate {
	
	private CompilationUnitEditor editor = null;

    /* (non-Javadoc)
     * @see org.eclipse.ui.IEditorActionDelegate#setActiveEditor(org.eclipse.jface.action.IAction, org.eclipse.ui.IEditorPart)
     */
    public void setActiveEditor(IAction action, IEditorPart targetEditor) {
		editor = (CompilationUnitEditor)targetEditor;
	}

    /* (non-Javadoc)
     * @see org.eclipse.ui.IActionDelegate2#runWithEvent(org.eclipse.jface.action.IAction, org.eclipse.swt.widgets.Event)
     */
    public void runWithEvent(IAction action, Event event) {
        ITextSelection sel = ((ITextSelection)editor.getSelectionProvider().getSelection());
        JavaTextSelection javaSelection= new JavaTextSelection(getEditorInput(), getDocument(), sel.getOffset(), sel.getLength());
        
        try {
            ASTNode selectedNode;
            ASTNode[] selectedNodes = javaSelection.resolveSelectedNodes();
            if (selectedNodes.length == 1) {
                selectedNode = selectedNodes[0];
            } else {
                selectedNode = javaSelection.resolveCoveringNode();
            }
            
            SelectedElement selectedElement = null;
            
            //could be more efficient, but it works now like this, so I prefered not to change it (hannes)
            IJavaElement[] elements = javaSelection.resolveElementAtOffset();  
            if (elements.length == 1 && isPossibleJavaElement(elements[0], selectedNode)) {
                if (elements[0].isReadOnly()) {
                    showError("Non-project type", "Declared type of selected declaration element is not part of this project."
                            + "\nPlease select a declaration element declared with a project type.");
                    return;
                } else {
                    selectedElement = getSelectedDeclarationElement(getTextSelection(elements[0]), elements[0]);
                }
            } else {
                selectedElement = getImportSelection(selectedNode);
                if (selectedElement != null) {
                    Scope scope = ImportScopeDialog.askForScope(getShell());
                    if (scope == null) {
                        return;
                    } else {
                        selectedElement.setScope(scope);
                    }
                } else {
                    selectedElement = getTypeArgumentSelection(selectedNode);                                                        
                }
                                
            }

            if (selectedElement == null) {
                showError("Selection invalid", "You must select a field, parameter, local variable, method, type argument, or single type import declaration.");                
            } else {
                runWithSelectedElement(selectedElement);                        
            }
            
        } catch (InternalCalculationException e) {
            showError("Selection invalid", "Infer Type cannot be performed on your selection.");
        } catch (JavaModelException e) {
            e.printStackTrace();
            showError("Selection invalid", "Could not resolve the selected element: " + e.getMessage());
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

    }
    
    private SelectedElement getImportSelection(ASTNode selectedNode) {
        
        ASTNode node = selectedNode;
        while (node != null) {
            if (node instanceof ImportDeclaration) {
                ImportDeclaration importDeclaration = (ImportDeclaration)node;
                IBinding binding = importDeclaration.resolveBinding();
                if (binding instanceof ITypeBinding) {
                    return new SelectedElement((ITypeBinding)binding, selectedNode);                    
                }
            }
            node = node.getParent();
        }
        return null;
    }
    
    private static ASTNode getUsedParent(ASTNode[] usedChild) {
        ASTNode parent = usedChild[0].getParent();
        
        if (parent instanceof WildcardType) {
            usedChild[0] = parent;
            parent = ((WildcardType)parent).getParent();
        }
        while (parent instanceof Type && ((Type)parent).isArrayType()) {
            usedChild[0] = parent;
            parent = ((Type)parent).getParent();
        }
        return parent;
    }

    private SelectedElement getTypeArgumentSelection(final ASTNode initialSelectedNode) {
        ASTNode[] selectedNode = new ASTNode[1];
        selectedNode[0] = initialSelectedNode;
        
        ASTNode parent = getUsedParent(selectedNode);

        if (selectedNode[0] instanceof SimpleName && parent instanceof Type)
        {
            selectedNode[0] = parent;
            parent = getUsedParent(selectedNode);
        }

        if (selectedNode[0] instanceof SimpleType && parent instanceof ParameterizedType && 
                ((ParameterizedType)parent).getType().equals(selectedNode[0]))
        {
            selectedNode[0] = parent;
            parent = getUsedParent(selectedNode);
        }
        
        List<String> typeVariableKeys = new ArrayList<String>();

        while (parent instanceof ParameterizedType) {
            
            ParameterizedType parameterizedType = (ParameterizedType)parent;
            ITypeBinding[] typeVariables =  parameterizedType.resolveBinding().getTypeDeclaration().getTypeParameters();
            List typeArguments = parameterizedType.typeArguments();
            for (int i = 0; i < typeVariables.length; i++) {
                if (typeArguments.get(i).equals(selectedNode[0])) {
                    typeVariableKeys.add(typeVariables[i].getKey());
                }                
            }
            selectedNode[0] = parent;
            parent = getUsedParent(selectedNode);
        }
        
        Collections.reverse(typeVariableKeys);
        
        SelectedElement selectedElement = null;
        if (!typeVariableKeys.isEmpty()) {
            if (parent instanceof FieldDeclaration) {
                FieldDeclaration fieldDeclaration = (FieldDeclaration)parent;
                IBinding binding = ((VariableDeclarationFragment)fieldDeclaration.fragments().get(0)).getName().resolveBinding();
                selectedElement = new SelectedElement(binding, 
                        fieldDeclaration.getType().getStartPosition(), 
                        typeVariableKeys);
            } else if (parent instanceof SingleVariableDeclaration) {
                SingleVariableDeclaration singleVariableDeclaration = (SingleVariableDeclaration)parent;
                selectedElement = new SelectedElement(singleVariableDeclaration.resolveBinding(), 
                        getDeclaringType(parent),
                        singleVariableDeclaration.getType().getStartPosition(), 
                        typeVariableKeys);
            } else if (parent instanceof VariableDeclarationStatement) {
                VariableDeclarationStatement variableDeclarationStatement = (VariableDeclarationStatement)parent;
                IBinding binding = ((VariableDeclarationFragment)variableDeclarationStatement.fragments().get(0)).getName().resolveBinding();
                selectedElement = new SelectedElement(binding, 
                        getDeclaringType(parent),
                        variableDeclarationStatement.getType().getStartPosition(), 
                        typeVariableKeys);
            } else if (parent instanceof MethodDeclaration) {
                MethodDeclaration methodDeclaration = (MethodDeclaration)parent;
                selectedElement = new SelectedElement(methodDeclaration.resolveBinding(), 
                        methodDeclaration.getReturnType2().getStartPosition(), 
                        typeVariableKeys);
            }
            
        }
        return selectedElement;
    }

    private static ITextSelection getTextSelection(final IJavaElement element) throws JavaModelException {
        if (element instanceof IType) {
            return null;
        }
        ISourceRange sourceRange = ((ISourceReference)element).getSourceRange();
        if (element instanceof IField) {
            sourceRange = ((IField)element).getNameRange();
        } else if (element instanceof IMethod) {
            sourceRange = ((IMethod)element).getNameRange();
        } else if (element instanceof ILocalVariable) {
            sourceRange = ((ILocalVariable)element).getNameRange();
        }
        return new TextSelection(sourceRange.getOffset(), sourceRange.getLength());
    }
    
    private SelectedElement getSelectedDeclarationElement(ITextSelection text, IJavaElement element) throws InternalCalculationException {
        ICompilationUnit icu = getCompilationUnit(element);
        
        ASTParser parser = ASTParser.newParser(AST.JLS3);
        parser.setSource(icu);
        parser.setResolveBindings(true);
        CompilationUnit compilationUnit = (CompilationUnit) parser.createAST(null);
        SelectedElementVisitor vis = new SelectedElementVisitor(text);
        compilationUnit.accept(vis);

        if (vis.getSelectedElement() == null) {
            throw new InternalCalculationException("Not a valid selection.\n"
                    + "You must select a reference (field, parameter, local variable, or method)");
        }

        SelectedElement declarationElement = vis.getSelectedElement();
        declarationElement.setCompilationUnit(icu);
        return declarationElement;
    }

    private ICompilationUnit getCompilationUnit(IJavaElement element) {
        IJavaElement icu = element;        
        while (!(icu instanceof ICompilationUnit)) {
            icu = icu.getParent();
        }
        return (ICompilationUnit)icu;
    }

    
    private ITypeRoot getEditorInput() {
        return JavaPlugin.getDefault().getWorkingCopyManager().getWorkingCopy(
                editor.getEditorInput());      
    }
    
    private IDocument getDocument() {
        return JavaPlugin.getDefault().getCompilationUnitDocumentProvider().
            getDocument(editor.getEditorInput());
    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IActionDelegate#run(org.eclipse.jface.action.IAction)
     */
    public void run(IAction action) {
    }
    
    /* (non-Javadoc)
     * @see org.eclipse.ui.IActionDelegate2#dispose()
     */
    public void dispose() {
    }

    /* (non-Javadoc)
     * @see org.eclipse.ui.IActionDelegate2#init(org.eclipse.jface.action.IAction)
     */
    public void init(IAction action) {       
    }

    private boolean isPossibleJavaElement(final IJavaElement element, ASTNode node) {
        if (element instanceof ISourceReference) {
            switch (element.getElementType()) {
                case IJavaElement.METHOD:
                case IJavaElement.FIELD:
                case IJavaElement.LOCAL_VARIABLE:
                    return true;
            }
        }
        return false;        
    }

    private static ITypeBinding getDeclaringType(ASTNode node) {

        ASTNode cnode = node;
        while (cnode != null) {
            if (cnode instanceof AbstractTypeDeclaration) {
                return ((AbstractTypeDeclaration) cnode).resolveBinding();
            }
            cnode = cnode.getParent();
        }
        return null;
    }    
}
