/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;

/**
 * The type constraint that reflects the type equality requirement of two
 * type variables.
 * <p>
 * In the JDT refactorings this is not needed because the two type variables
 * get the same ITypeSet. This is not possible for us. In addition, the space
 * saving would be really small and the explicit equality constraint is better for 
 * debugging.
 * </p>
 *  
 * @author Hannes Kegel
 *
 */
public class EqualityTypeConstraint implements ITypeConstraint2 {

    private final ConstraintVariable2 right;
    private final ConstraintVariable2 left;

    /**
     * The single constructor. The order of the variables doesn't matter.
     * @param left the first variable
     * @param right the second variable
     */
    public EqualityTypeConstraint(final ConstraintVariable2 left, final ConstraintVariable2 right) {
        Assert.isNotNull(left);
        Assert.isNotNull(right);
        this.left= left;
        this.right= right;
    }

    /*
     * @see java.lang.Object#equals(java.lang.Object)
     */
    @Override
    public final boolean equals(Object other) {
        if (other.getClass() != EqualityTypeConstraint.class) {
            return false;
        }

        // can use object identity on ConstraintVariables, since we have the stored (or to be stored) objects
        ITypeConstraint2 otherTC= (ITypeConstraint2) other;
        return (left == otherTC.getLeft() && right == otherTC.getRight()) ||
                (right == otherTC.getLeft() && left == otherTC.getRight());
    }

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2#getLeft()
     */
    public final ConstraintVariable2 getLeft() {
        return left;
    }

    /* (non-Javadoc)
     * @see org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2#getRight()
     */
    public final ConstraintVariable2 getRight() {
        return right;
    }

    /*
     * @see java.lang.Object#hashCode()
     */
    @Override
    public final int hashCode() {
        return left.hashCode() ^ right.hashCode();
    }

    /*
     * @see java.lang.Object#toString()
     */
    @Override
    public final String toString() {
        return left.toString() + " == " + right.toString(); //$NON-NLS-1$ 
    }
    
}
