/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.creation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.WorkingCopyOwner;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.ASTRequestor;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.IBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.SearchPattern;
import org.eclipse.jdt.internal.corext.refactoring.Checks;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringScopeFactory;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringSearchEngine2;
import org.eclipse.jdt.internal.corext.refactoring.SearchResultGroup;
import org.eclipse.jdt.internal.corext.refactoring.structure.ASTNodeSearchUtil;
import org.eclipse.jdt.internal.corext.util.SearchUtils;
import org.eclipse.jdt.internal.ui.JavaPlugin;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.InferTypeRefactoring;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.util.SelectedElement;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.InternalCalculationException;
import org.intoJ.inferType3.internal.util.LogUtil;

/**
 * The Collector searches for all affected compilation units and
 * feeds them to the {@link InferTypeConstraintsCreator}.
 * The action starts with the run method.
 * <p>
 * First, it searches for the occurences of all types in the hierarchy and the cus
 * containing them will be parsed (this is called the first pass). 
 * Then all types where the initial types are used in extends or implements
 * clauses as arguments of parameterized types (like MyComparator implements Comparator<Test>)
 * are added to the search, too. (the second pass).
 * After that all usages of methods that return one of these types or fields typed like this
 * will be searched (the third pass).
 * </p>
 * 
 * @author Hannes Kegel
 *
 */
//TODO: IT4: optimize

public class Collector {
    /** Number of compilation units to parse at once */
    private static final int SIZE_BATCH = 500;

    private final WorkingCopyOwner owner;
    private final IJavaProject project;
    private final InferTypeConstraintsModel model;
    private final TypeHierarchy typeHierarchy;
    private final boolean includeThisReceiver;
    private final boolean includeThisReceiverForSubclassFields;
    private final boolean includeSuperMethodInvocations;
    private final boolean includeCasts;
    private final boolean parseDeclarations;

	private final boolean includeInstanceOf;

	
    /**
     * The single constructor.
     * 
     * @param owner the working copy owner
     * @param model the model that will be filled
     * @param project the java project
     * @param typeHierarchy the type hierarchy that we will search for
     */
    public Collector(WorkingCopyOwner owner, InferTypeConstraintsModel model, IJavaProject project, TypeHierarchy typeHierarchy, boolean includeThisReceiver, boolean includeThisReceiverForSubclassFields, boolean includeSuperMethodInvocations, boolean includeCasts, boolean parseDeclarations, boolean includeInstanceOf) {
        this.owner = owner;
        this.model = model;
        this.typeHierarchy = typeHierarchy;
        this.project = project;
        this.includeThisReceiver = includeThisReceiver;
		this.includeThisReceiverForSubclassFields = includeThisReceiverForSubclassFields;
		this.includeSuperMethodInvocations = includeSuperMethodInvocations;
		this.includeCasts = includeCasts;
		this.parseDeclarations = parseDeclarations;
		this.includeInstanceOf = includeInstanceOf;
    }

    /**
     * The entry point.
     * 
     * @param monitor the current progress monitor
     * @param status the current refactoring status used to return errors
     */
    public void run(final IProgressMonitor monitor, final RefactoringStatus status) {
        final ASTParser parser = ASTParser.newParser(AST.JLS3);
        parser.setWorkingCopyOwner(owner);
        parser.setResolveBindings(true);
        parser.setProject(project);
        parser.setCompilerOptions(Cache.getCompilerOptions(project));

        
        final ITypeBinding declaredTypeBinding = typeHierarchy.getOriginalSelectedType().getBaseTypeBinding();;
        final IType declaredType = (IType)declaredTypeBinding.getJavaElement();

        parser.createASTs(new ICompilationUnit[] { (ICompilationUnit) declaredTypeBinding.getJavaElement().getAncestor(IJavaElement.COMPILATION_UNIT) }, new String[0], new ASTRequestor() {
            @Override
            public final void acceptAST(final ICompilationUnit unit, final CompilationUnit node) {
                try {
                    buildConstraints(unit, node, declaredType, declaredTypeBinding, monitor, status);
                } catch (InternalCalculationException e) {
                    status.addFatalError(e.getMessage());
                } catch (CoreException exception) {
                    JavaPlugin.log(exception);
                    status.merge(RefactoringStatus.createFatalErrorStatus(InferTypeRefactoring.MSG_INTERNAL_ERROR));
                }
            }

            @Override
            public final void acceptBinding(final String key, final IBinding binding) {
                // Do nothing
            }
        }, new NullProgressMonitor());
    }
    

    private final void buildConstraints(final ICompilationUnit cuOfSelection, final CompilationUnit cuNodeOfSelection, final IType declaredType, final ITypeBinding declaredBinding,
            final IProgressMonitor monitor, final RefactoringStatus status) throws JavaModelException, InternalCalculationException {

        final InferTypeConstraintsCreator creator = new InferTypeConstraintsCreator(model, includeThisReceiver, includeThisReceiverForSubclassFields, includeSuperMethodInvocations, includeCasts, includeInstanceOf, typeHierarchy, status);
        final ASTParser parser = ASTParser.newParser(AST.JLS3);
        final Set<ICompilationUnit> processed = new HashSet<ICompilationUnit>();

        try {
            monitor.beginTask("", 100);
            monitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS);
            
            // parse all files where types are used in three passes
            final Map<IJavaProject, Set<ICompilationUnit>> thirdPass = startFirstPass(typeHierarchy, cuOfSelection, cuNodeOfSelection, parser, declaredType, processed, creator, new SubProgressMonitor(monitor, 75), status);

            startThirdPass(new SubProgressMonitor(monitor, 25), creator, parser, processed, thirdPass, status);

        } finally {
            monitor.done();
        }
    }
    
    private Map<IJavaProject, Set<ICompilationUnit>> startFirstPass(TypeHierarchy typeHierarchy, final ICompilationUnit cuOfSelection, final CompilationUnit cuNodeOfSelection, final ASTParser parser,
            final IType declaredType, final Set<ICompilationUnit> processed, final InferTypeConstraintsCreator creator, final IProgressMonitor monitor, final RefactoringStatus status)
            throws JavaModelException {
        final Map<IJavaProject, Set<ICompilationUnit>> thirdPassUnits = new HashMap<IJavaProject, Set<ICompilationUnit>>();

        final RefactoringSearchEngine2 engine = createEngine(declaredType, status);
        
        monitor.beginTask("", 120);
        monitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS);

        final Map<IJavaProject, Collection<SearchResultGroup>> initialFirstPass = getReferencingCompilationUnits(
                CollectorUtil.getITypesFromHierarchy(typeHierarchy), new SubProgressMonitor(monitor, 20, SubProgressMonitor.SUPPRESS_SUBTASK_LABEL), engine);
        Map<ICompilationUnit, SearchResultGroup> groups = CollectorUtil.getCompUnitToGroups(initialFirstPass);
        Set<IType> typesUsingOtherInParametricDeclaration = new HashSet<IType>();

        if (cuOfSelection != null) {
            processed.add(cuOfSelection);
        }
        
        IProgressMonitor subMonitor = new SubProgressMonitor(monitor, 70);
        try {
            subMonitor.beginTask("", initialFirstPass.size() * 100);
            subMonitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS); 
            
            for (Map.Entry<IJavaProject, Collection<SearchResultGroup>> entry : initialFirstPass.entrySet()) {
                Collection<SearchResultGroup> srGroups = entry.getValue();
                if (srGroups != null) {
                    Set<ICompilationUnit> unitsToBeParsed = CollectorUtil.getUnitsToBeParsed(srGroups, processed);

                    for (ICompilationUnit unit : unitsToBeParsed) {
                        LogUtil.debugOut(this, "FP PARSING UNIT " + unit.getElementName());
                        Checks.checkCompileErrorsInAffectedFile(status, unit.getResource());            
                    }

                    parseGroup(parser, processed, creator, thirdPassUnits, engine, typesUsingOtherInParametricDeclaration, groups, subMonitor, entry.getKey(), unitsToBeParsed);
                }
            }
        } finally {
            initialFirstPass.clear();
            subMonitor.done();
        }
        if (cuOfSelection != null && cuNodeOfSelection != null) {
            Checks.checkCompileErrorsInAffectedFile(status, cuOfSelection.getResource());            
            performFirstPass(creator, typesUsingOtherInParametricDeclaration, thirdPassUnits, groups, cuOfSelection, cuNodeOfSelection, engine, new SubProgressMonitor(monitor, 5));
        } else {
            monitor.worked(5);
        }

        startSecondPass(parser, processed, creator, monitor, status, thirdPassUnits, engine, typesUsingOtherInParametricDeclaration);
        monitor.done();
        
        return thirdPassUnits;

    }

    // this is done for references to classes where our type is used in the declaration
    // like class MyComparator implements Comparator<TheTypeInQuestion>
    private void startSecondPass(final ASTParser parser, final Set<ICompilationUnit> processed, final InferTypeConstraintsCreator creator, final IProgressMonitor monitor,
            final RefactoringStatus status, final Map<IJavaProject, Set<ICompilationUnit>> thirdPassUnits, final RefactoringSearchEngine2 engine, Set<IType> typesUsingOtherInParametricDeclaration)
            throws JavaModelException {

        IProgressMonitor secondPassMonitor = new SubProgressMonitor(monitor, 25);

        Set<IType> allTypesUsingOtherInParametricDeclaration = new HashSet<IType>();
        while (!typesUsingOtherInParametricDeclaration.isEmpty()) {
            // we have to reparse affected files
            Set<ICompilationUnit> processedNow = new HashSet<ICompilationUnit>();
            
            Map<IJavaProject, Collection<SearchResultGroup>> additionalFirstPass = getReferencingCompilationUnits(
                    typesUsingOtherInParametricDeclaration, new SubProgressMonitor(monitor, 20, SubProgressMonitor.SUPPRESS_SUBTASK_LABEL), engine);
            
            Set<IType> nextTypesUsingOtherInParametricDeclaration = new HashSet<IType>();
            allTypesUsingOtherInParametricDeclaration.addAll(typesUsingOtherInParametricDeclaration);
            
            Map<ICompilationUnit, SearchResultGroup> groups = CollectorUtil.getCompUnitToGroups(additionalFirstPass);
            for (Map.Entry<IJavaProject, Collection<SearchResultGroup>> entry : additionalFirstPass.entrySet()) {
                Collection<SearchResultGroup> srGroups = entry.getValue();
                if (srGroups != null) {
                    Set<ICompilationUnit> unitsToBeParsed = CollectorUtil.getUnitsToBeParsed(srGroups, processedNow);

                    for (ICompilationUnit unit : unitsToBeParsed) {
                        LogUtil.debugOut(this, "SP PARSING UNIT " + unit.getElementName());
                        Checks.checkCompileErrorsInAffectedFile(status, unit.getResource());            
                    }

                    parseGroup(parser, processedNow, creator, thirdPassUnits, engine, nextTypesUsingOtherInParametricDeclaration, groups, secondPassMonitor, entry.getKey(), unitsToBeParsed);
                }
            }
            processed.addAll(processedNow);
            nextTypesUsingOtherInParametricDeclaration.removeAll(allTypesUsingOtherInParametricDeclaration);
            typesUsingOtherInParametricDeclaration = nextTypesUsingOtherInParametricDeclaration;
        }
        secondPassMonitor.done();
    }

    private void parseGroup(final ASTParser parser, final Set<ICompilationUnit> processed, 
            final InferTypeConstraintsCreator creator, final Map<IJavaProject, 
            Set<ICompilationUnit>> thirdPassUnits,
            final RefactoringSearchEngine2 engine, 
            final Set<IType> typesUsingOtherInParametricDeclaration, 
            final Map<ICompilationUnit, SearchResultGroup> groups, IProgressMonitor monitor, 
            IJavaProject project, Set<ICompilationUnit> unitsToBeParsed) 
    {
        final List<ICompilationUnit> batches = new ArrayList<ICompilationUnit>(unitsToBeParsed);
        final int size = batches.size();
        final int iterations = ((size - 1) / SIZE_BATCH) + 1;
        final IProgressMonitor subMonitor = new SubProgressMonitor(monitor, 100);
        try {
            subMonitor.beginTask("", iterations * 100); 
            subMonitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS); 
            
            final Map options = Cache.getCompilerOptions(project);
            for (int index = 0; index < iterations; index++) {
                final List<ICompilationUnit> iteration = batches.subList(index * SIZE_BATCH, Math.min(size, (index + 1) * SIZE_BATCH));
                parser.setWorkingCopyOwner(owner);
                parser.setResolveBindings(true);
                parser.setProject(project);
                parser.setCompilerOptions(options);
                final IProgressMonitor subSubMonitor = new SubProgressMonitor(subMonitor, 100);
                try {
                    final int count = iteration.size();
                    subSubMonitor.beginTask("", count * 100); 
                    subSubMonitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS); 
                    parser.createASTs(iteration.toArray(new ICompilationUnit[count]), new String[0], new ASTRequestor() {

                        @Override
                        public final void acceptAST(final ICompilationUnit unit, final CompilationUnit node) {
                            if (!processed.contains(unit)) {
                                performFirstPass(creator, typesUsingOtherInParametricDeclaration, thirdPassUnits, groups, unit, node, engine, new SubProgressMonitor(subSubMonitor, 100));
                                processed.add(unit);
                            } else {
                                subSubMonitor.worked(100);
                            }
                        }

                        @Override
                        public final void acceptBinding(final String key, final IBinding binding) {
                            // Do nothing
                        }
                    }, new NullProgressMonitor());
                } finally {
                    subSubMonitor.done();
                }
            }
        } finally {
            subMonitor.done();
        }
    }

    private RefactoringSearchEngine2 createEngine(final IType declaredType, final RefactoringStatus status) throws JavaModelException {
        final RefactoringSearchEngine2 engine = new RefactoringSearchEngine2();
        engine.setOwner(owner);
        engine.setFiltering(true, true);
        engine.setStatus(status);
        engine.setScope(RefactoringScopeFactory.create(declaredType, false));
        return engine;
    }

    private void startThirdPass(final IProgressMonitor monitor, final InferTypeConstraintsCreator creator, final ASTParser parser,
            final Set<ICompilationUnit> processed, final Map<IJavaProject, Set<ICompilationUnit>> thirdPassUnits, RefactoringStatus status) throws JavaModelException {
        try {
            try {
                final Set<IJavaProject> keySet = thirdPassUnits.keySet();
                monitor.beginTask("", keySet.size() * 100);
                monitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS); 
                
                for (IJavaProject project : keySet) {
                    Collection<ICompilationUnit> unitsToBeParsed = thirdPassUnits.get(project);
                    if (unitsToBeParsed != null) {
                        for (ICompilationUnit unit : unitsToBeParsed) {
                            if (!processed.contains(unit)) {
                                LogUtil.debugOut(this, "TP PARSING UNIT " + unit.getElementName());
                                Checks.checkCompileErrorsInAffectedFile(status, unit.getResource());
                            }
                        }
                        parser.setWorkingCopyOwner(owner);
                        parser.setResolveBindings(true);
                        parser.setProject(project);
                        parser.setCompilerOptions(Cache.getCompilerOptions(project));
                        final IProgressMonitor subsubMonitor = new SubProgressMonitor(monitor, 100);
                        try {
                            subsubMonitor.beginTask("", unitsToBeParsed.size() * 100); 
                            subsubMonitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS);
                            
                            parser.createASTs(unitsToBeParsed.toArray(new ICompilationUnit[unitsToBeParsed.size()]), new String[0], new ASTRequestor() {

                                @Override
                                public final void acceptAST(final ICompilationUnit unit, final CompilationUnit node) {
                                    if (!processed.contains(unit)) {
                                        node.accept(creator);
                                    }
                                    subsubMonitor.worked(100);
                                }

                                @Override
                                public final void acceptBinding(final String key, final IBinding binding) {
                                    // Do nothing
                                }
                            }, new NullProgressMonitor());
                        } finally {
                            subsubMonitor.done();
                        }
                    }
                }
            } finally {
                thirdPassUnits.clear();
            }
        } finally {
            creator.endCreation();
            monitor.done();
        }
    }
    

    /**
     * Computes the compilation units referencing the types to replace.
     * 
     * @param types
     *            the types to search for
     * @param monitor
     *            the progress monitor to use
     * @param engine
     *            the search engine to used
     * @return the referenced compilation units 
     * @throws JavaModelException
     */
    @SuppressWarnings("unchecked")
    private Map<IJavaProject, Collection<SearchResultGroup>> getReferencingCompilationUnits(Collection<IType> types, final IProgressMonitor monitor,
            final RefactoringSearchEngine2 engine) throws JavaModelException {
        try {
            monitor.beginTask("", 100); 
            monitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS); 

            SearchPattern completePattern = null;
            for (IType type : types) {
                LogUtil.debugOut(Collector.class, "SEARCHING FOR " + type.getFullyQualifiedName());
                if (!type.getFullyQualifiedName().equals("java.lang.Object")) {
                    SearchPattern singlePattern = SearchPattern.createPattern(type, IJavaSearchConstants.REFERENCES, SearchUtils.GENERICS_AGNOSTIC_MATCH_RULE);
                    if (completePattern == null) {
                        completePattern = singlePattern;
                    } else {
                        completePattern = SearchPattern.createOrPattern(completePattern, singlePattern);
                    }
                    if (parseDeclarations && !type.isInterface()) {
                        singlePattern = SearchPattern.createPattern(type, IJavaSearchConstants.DECLARATIONS, SearchUtils.GENERICS_AGNOSTIC_MATCH_RULE);
                        completePattern = SearchPattern.createOrPattern(completePattern, singlePattern);                    	
                    }
                }
            }
                        
            engine.setPattern(completePattern);
            engine.searchPattern(new SubProgressMonitor(monitor, 100));
            return engine.getAffectedProjects();
        } finally {
            monitor.done();
        }
    }

    /**
     * Performs the first pass of processing the affected compilation units.
     * 
     * @param creator
     *            the constraints creator to use
     * @param thirdPassUnits
     *            the compilation unit map (element type:
     *            <code>&ltIJavaProject, Set&ltICompilationUnit&gt&gt</code>)
     * @param groups
     *            the search result group map (element type:
     *            <code>&ltICompilationUnit, SearchResultGroup&gt</code>)
     * @param unit
     *            the compilation unit of the subtype
     * @param node
     *            the compilation unit node of the subtype
     * @param monitor
     *            the progress monitor to use
     */
    private void performFirstPass(final InferTypeConstraintsCreator creator, 
            final Set<IType> typesUsingOtherInParametricDeclaration, 
            final Map<IJavaProject, Set<ICompilationUnit>> thirdPassUnits, 
            final Map<ICompilationUnit, SearchResultGroup> groups,
            final ICompilationUnit unit, final CompilationUnit node, RefactoringSearchEngine2 engine, 
            final IProgressMonitor monitor) 
    {
        try {
            monitor.beginTask("", 100);
            monitor.setTaskName(InferTypeRefactoring.MSG_CALCULATING_ACCESS_SETS); 
            
            node.accept(creator);
            monitor.worked(20);
            final SearchResultGroup group = groups.get(unit);
            if (group != null) {
                final ASTNode[] nodes = ASTNodeSearchUtil.getAstNodes(group.getSearchResults(), node);
                try {
                    CollectorUtil.getMethodReferencingCompilationUnits(thirdPassUnits, nodes, engine);
                    monitor.worked(35);
                    CollectorUtil.getFieldReferencingCompilationUnits(thirdPassUnits, nodes, engine);
                    monitor.worked(35);
                    CollectorUtil.getDeclarationReferencingCompilationUnits(typesUsingOtherInParametricDeclaration, nodes);
                    monitor.worked(10);
                } catch (JavaModelException exception) {
                    JavaPlugin.log(exception);
                }
            }
        } finally {
            monitor.done();
        }
    }
    

}
