/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.util;

import java.io.File;
import java.io.IOException;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.SubTypeConstraint2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.TypeVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.VariableVariable2;
import org.intoJ.inferType3.internal.constraintModel.EqualityTypeConstraint;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.variables.DeclaringTypeVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;
import org.intoJ.inferType3.internal.util.GmlWriter;

/**
 * Writes the model to a {@link GmlWriter}. See this class for more information
 * about the file format and how it can be opened.
 * 
 * @author Hannes Kegel
 *
 */
public class InferTypeGmlHandler {
    
    private static final boolean GERMAN_TEXT = Boolean.getBoolean("inferType.german");

    /**
     * Exports the given model to a gml file. See {@link GmlWriter} for file format
     * hints. 
     * 
     * @param path the desired file path 
     * @param writeElementEdges if the graph should contain edges that show the relation between element 
     * variables and its parametrized types 
     * @param onlyTraversed if only the part of the graph that will be traversed for IT should be exported
     * @param model the model to export
     * @param hierarchy the type hierarchy
     * @throws IOException
     * @throws JavaModelException
     */
    public static void writeGml(String path, boolean writeElementEdges, boolean onlyTraversed, InferTypeConstraintsModel model, TypeHierarchy hierarchy) throws IOException, JavaModelException {
        if (!path.toLowerCase().endsWith(".gml")) {
            path = path + ".gml";
        }
        
        Collection<ConstraintVariable2> constraintVariables = model.getConstraintVariables();        
        Collection<ITypeConstraint2> allConstraints = model.getTypeConstraints();
        ConstraintVariable2 selectedVariable = model.getSelectedConstraintVariable();
        
        GmlWriter gmlWriter = new GmlWriter(new File(path));
        
        if (onlyTraversed) {
            final Set<ITypeConstraint2> constraints = new HashSet<ITypeConstraint2>();
            final Set<ConstraintVariable2> variables = new HashSet<ConstraintVariable2>();
            if (selectedVariable != null) {
                ConstraintTraversal.visitAllDirectedConstraints(selectedVariable, new ConstraintTraversal.Visitor() {

                    public boolean visit(ConstraintVariable2 baseConstraintVariable, ConstraintVariable2 otherConstraintVariable2, ITypeConstraint2 constraint) throws JavaModelException {
                        constraints.add(constraint);
                        variables.add(baseConstraintVariable);
                        variables.add(otherConstraintVariable2);
                        return true;
                    }

                });

                write(writeElementEdges, selectedVariable, gmlWriter, constraints, variables, hierarchy);
            }
        } else {        
            write(writeElementEdges, selectedVariable, gmlWriter, allConstraints, constraintVariables, hierarchy);
        }

        if (!gmlWriter.finish()) {
            throw new IOException("Could not write GML to file " + path);
        }

    }
    
    /**
     * Does the same as writeGml without throwing exceptions. This is useful for debugging.
     * 
     * @param path the desired file path 
     * @param writeElementEdges if the graph should contain edges that show the relation between element 
     * variables and its parametrized types 
     * @param onlyTraversed if only the part of the graph that will be traversed for IT should be exported
     * @param model the model to export
     * @param hierarchy the type hierarchy
     */
    public static void writeGmlNoEx(String path, boolean writeElementEdges, boolean onlyTraversed, InferTypeConstraintsModel model, TypeHierarchy hierarchy) {
        try {
            writeGml(path, writeElementEdges, onlyTraversed, model, hierarchy);
        } catch (JavaModelException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    

    private static void write(boolean writeElementEdges, ConstraintVariable2 selectedVariable, GmlWriter gmlWriter,
            final Collection<ITypeConstraint2> constraints, final Collection<ConstraintVariable2> variables, TypeHierarchy hierarchy) throws JavaModelException 
    {
        for (ConstraintVariable2 var : variables) {
            writeGmlVariable(gmlWriter, var, selectedVariable, hierarchy);
        }

        for (ITypeConstraint2 constraint : constraints) {
            writeGmlConstraint(gmlWriter, constraint);
        }

        if (writeElementEdges) {
            for (ConstraintVariable2 var : variables) {
                for (Map.Entry<String, ElementVariable> entry : InferTypeConstraintsModel.getElementVariables(var).entrySet()) {
                    gmlWriter.writeEdge(var, entry.getValue(), getPrettyTypeVariable(entry.getKey()), "#C0C0C0", null, true);
                }
            }
        }
    }

    private static void writeGmlConstraint(GmlWriter gmlWriter, ITypeConstraint2 constraint) {
        if (constraint instanceof EqualityTypeConstraint) {
            gmlWriter.writeEdge(constraint.getLeft(), constraint.getRight(), null, null, "dashed", false);
        } else if (constraint instanceof SubTypeConstraint2) {
            gmlWriter.writeEdge(constraint.getLeft(), constraint.getRight(), null, null, "solid", true);
        }
    }

    private static void writeGmlVariable(GmlWriter gmlWriter, ConstraintVariable2 var, ConstraintVariable2 selectedVariable, TypeHierarchy hierarchy) throws JavaModelException {
        String color;
        if (selectedVariable == var) {
            color = "#607BBF";
        } else if (var instanceof DeclaringTypeVariable) {
            color = "#80A4FF"; // "#D0D4FC"; 
        } else {
            color = "#DBE5F1";
        }
        gmlWriter.writeNode(var, getVariableTypeName(var, hierarchy), color);
    }

    private static String getVariableTypeName(ConstraintVariable2 var, TypeHierarchy hierarchy) {
        String simpleName = InferTypeConstraintsModel.getSimpleName(var);
        TCType newType = InferTypeConstraintsModel.getNewType(var);

        StringBuilder ret = new StringBuilder();
        
        if (GERMAN_TEXT) {
            if (var instanceof DeclaringTypeVariable) {
                DeclaringTypeVariable declaringTypeVariable = (DeclaringTypeVariable)var;
                BindingWrapper wrapper = hierarchy.getMethod(declaringTypeVariable.getBindingKey());
                ret.append("Decl(");
                if (wrapper == null) {
                    ret.append(declaringTypeVariable.getBindingKey());
                } else {
                    IMethodBinding methodBinding = wrapper.getMethodBinding();
                    ret.append(methodBinding.getDeclaringClass().getName()).append(".") 
                            .append(methodBinding.getName()) 
                            .append(methodBinding.getParameterTypes().length == 0 ? "()" : "(...)");
                }
                ret.append(")");
            } else if (var instanceof TypeVariable2){
                if (simpleName == null) {
                    ret.append("Konstruktor");                    
                }
            } else if (var instanceof ElementVariable){                
                ret.append(getElementVariableText((ElementVariable)var));
            } else if (!(var instanceof VariableVariable2)) {
                ret.append(var.toString()).append("\n").append(var.getClass().getSimpleName());
            }
            if (simpleName != null) {
                ret.append("[").append(simpleName).append("]");
            }
            
            if (var.getType() != null) {
                ret.append("\nStartbel.: ").append(var.getType().getName());
            }
            if (newType != null) {
                ret.append("\nEndbel.: ").append(newType.getSimpleName());
            }
            //ret.append("\n id: " +System.identityHashCode(InferTypeConstraintsModel.getAccessSet(var)) + ": " + InferTypeConstraintsModel.getAccessSet(var));
        } else {
            ret.append(var.toString()).append("\n").append(var.getClass().getSimpleName());
            ret.append("\ntype: ").append(var.getType());
            if (simpleName != null) {
                ret.append(": ").append(simpleName);
            }
            if (newType != null) {
                ret.append("\nnew type: ").append(newType);
            }            
        }

        return ret.toString();
    }

    private static String getElementVariableText(ElementVariable var) {
        String parentText;
        ConstraintVariable2 parentCv = var.getParentConstraintVariable();
        if (parentCv instanceof ElementVariable) {
            parentText = getElementVariableText((ElementVariable)parentCv);
        } else {
            parentText = parentCv.getType().getName();            
        }
        return "Elem(" + parentText + ", " + getPrettyTypeVariable(var.getTypeVariableKey()) + ")"; 
    }

    private static String getPrettyTypeVariable(String typeVariableKey) {
        int typeEnd = typeVariableKey.indexOf(';');
        String className = typeVariableKey.substring(1, typeEnd).replace('/', '.');
        if (GERMAN_TEXT) {
            className = className.substring(className.lastIndexOf(".") + 1);
        }
        return  className + ":" + typeVariableKey.substring(typeEnd + 3, typeVariableKey.length() - 1);
    }

}
