/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.constraintModel.util;

import java.util.Iterator;
import java.util.List;

import org.eclipse.jdt.core.dom.ASTNode;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.AbstractTypeDeclaration;
import org.eclipse.jdt.core.dom.EnhancedForStatement;
import org.eclipse.jdt.core.dom.FieldDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.SimpleName;
import org.eclipse.jdt.core.dom.SingleVariableDeclaration;
import org.eclipse.jdt.core.dom.Type;
import org.eclipse.jdt.core.dom.VariableDeclarationFragment;
import org.eclipse.jdt.core.dom.VariableDeclarationStatement;
import org.eclipse.jface.text.ITextSelection;

/**
 * This visitor determines the corresponding SelectedElement to a given offset and node length
 * of the type node.
 * 
 * @author Andreas Mei&szlig;ner
 * @author Hannes Kegel
 */
public class SelectedElementVisitor extends ASTVisitor {

    private SelectedElement selectedElement;
    private final int typeStartPosition;
    private final int typeLength;

    /**
     * A constructor for an ITextSelection.
     * 
     * @param text ITextSelection
     */
    public SelectedElementVisitor(ITextSelection text) {
        this(text.getOffset(), text.getLength());
    }

    /**
     * A constructor for an offset and length.
     *
     * @param typeStartPosition
     * @param typeLength
     */
    public SelectedElementVisitor(int typeStartPosition, int typeLength) {
        this.typeStartPosition = typeStartPosition;
        this.typeLength = typeLength;
    }

    /**
     * Returns the selected element
     * 
     * @return SelectedElement
     */
    public SelectedElement getSelectedElement() {

        return selectedElement;
    }

    /**
     * Visit a method declaration.
     * 
     * @param node MethodDeclaration
     * @return boolean
     */
    @Override
    public boolean visit(MethodDeclaration node) {

        if (selectedElement != null) {
            return true;
        }
        if (node == null) {
            return true;
        }

        SimpleName name = node.getName();
        if (isSelected(name)) {
            selectedElement = new SelectedElement(node.resolveBinding(), node.getReturnType2().getStartPosition(), null);
            return true;
        }

        int index = 0;
        List parameters = node.parameters();
        for (Iterator i = parameters.iterator(); i.hasNext();) {
            SingleVariableDeclaration singleVariableDeclaration = ((SingleVariableDeclaration) i.next());
            determineBinding(singleVariableDeclaration, index, node.resolveBinding());
            index++;
        }

        return true;
    }

    

    /**
     * Visit fields and local variables (here also for variables in for loops)
     * 
     * @param node VariableDeclarationFragment
     * @return boolean
     */
    @Override
    public boolean visit(VariableDeclarationFragment node) {
        if (selectedElement != null || node == null) {
            return true;
        }
        SimpleName name = node.getName();
        if (isSelected(name)) {
            selectedElement = new SelectedElement(name.resolveBinding(), getDeclaringType(name), getTypeStartPos(node.getParent()), null);
        }            
        return true;
    }
    
    

    /* (non-Javadoc)
     * @see org.eclipse.jdt.core.dom.ASTVisitor#endVisit(org.eclipse.jdt.core.dom.EnhancedForStatement)
     */
    @Override
    public void endVisit(EnhancedForStatement node) {
        SingleVariableDeclaration decl = node.getParameter();
        if (decl != null) {
            SimpleName name = decl.getName();
            if (isSelected(name)) {
                selectedElement = new SelectedElement(name.resolveBinding(), getDeclaringType(name), decl.getType().getStartPosition(), null);
            }
        }
        
    }

    private int getTypeStartPos(ASTNode parent) {
        if (parent instanceof VariableDeclarationStatement) {
            VariableDeclarationStatement decl = (VariableDeclarationStatement)parent;
            Type type = decl.getType();
            if (type != null) {
                return type.getStartPosition();
            }                   
        } else  if (parent instanceof FieldDeclaration) {
            FieldDeclaration decl = (FieldDeclaration)parent;
            Type type = decl.getType();
            if (type != null) {
                return type.getStartPosition();
            }                   
        }
        return 0;
    }
    
    private void determineBinding(SingleVariableDeclaration decl, int index, IMethodBinding method) {

        SimpleName name = decl.getName();
        if (isSelected(name)) {
            selectedElement = new SelectedElement(name.resolveBinding(), getDeclaringType(name), decl.getType().getStartPosition(), null);
        }
    }

    private boolean isSelected(ASTNode node) {
        return (node.getStartPosition() == typeStartPosition && node.getLength() == typeLength);
    }
    
    private static ITypeBinding getDeclaringType(ASTNode node) {

        ASTNode cnode = node;
        while (cnode != null) {
            if (cnode instanceof AbstractTypeDeclaration) {
                return ((AbstractTypeDeclaration) cnode).resolveBinding();
            }
            cnode = cnode.getParent();
        }
        return null;
    }    
    

}    
    
