/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.manipulation;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.WorkingCopyOwner;
import org.eclipse.jdt.core.dom.AST;
import org.eclipse.jdt.core.dom.ASTParser;
import org.eclipse.jdt.core.dom.ASTRequestor;
import org.eclipse.jdt.core.dom.ASTVisitor;
import org.eclipse.jdt.core.dom.AnonymousClassDeclaration;
import org.eclipse.jdt.core.dom.CompilationUnit;
import org.eclipse.jdt.core.dom.EnumDeclaration;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.MethodDeclaration;
import org.eclipse.jdt.core.dom.TypeDeclaration;
import org.eclipse.jdt.internal.corext.refactoring.changes.CompilationUnitChange;
import org.eclipse.jdt.internal.corext.refactoring.structure.CompilationUnitRewrite;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;

/**
 * This helper class is used after the type occurences have been updated and the new types
 * have been created (both not committed yet). If the changed occurences modify an extracted
 * method, the new method binding will be found and used instead of the old one.
 *  
 * @author Hannes Kegel
 *
 */
class BindingUpdater {

    /**
     * Checks whether type occurences have changed in extracted methods and replaces the
     * relevant bindings.
     *   
     * @param typeCreators the list of the creators of new types
     * @param typeHierarchy the type hierarchy of the selection
     * @param project the java project of the selection
     * @param unitToRewrite map from changed cu to the rewrite helper classes
     * @return true if a binding has been updated and thus the new types have to be updated, too.
     * @throws CoreException
     */
    public static boolean updateMethodBindings(List<TypeCreator> typeCreators, TypeHierarchy typeHierarchy, IJavaProject project, Map<ICompilationUnit, CompilationUnitRewrite> unitToRewrite) throws CoreException {
        WorkingCopyOwner owner = new WorkingCopyOwner() { };
        
        List<ICompilationUnit> newWcs = new ArrayList<ICompilationUnit>();
        for (TypeCreator creator : typeCreators) {
            ICompilationUnit wc = creator.getCompilationUnit().getWorkingCopy(owner, null, null);
            wc.getBuffer().setContents(creator.getContent());
            newWcs.add(wc);
        }
        
        
        Set<ICompilationUnit> existingCus = new HashSet<ICompilationUnit>();
        for (TCType tcType : typeHierarchy.getAllTypes()) {
            if (tcType instanceof ExistingTCType) {
                if (existingCus.add(((ExistingTCType)tcType).getIType().getCompilationUnit())) {
                    CompilationUnitRewrite rewrite = unitToRewrite.get(((ExistingTCType)tcType).getIType().getCompilationUnit());                    
                    if (rewrite != null) {
                        CompilationUnitChange cuChange = rewrite.createChange();
                        String previewContent = cuChange.getPreviewContent(null);

                        ICompilationUnit wc = rewrite.getCu().getWorkingCopy(owner, null, null);
                        wc.getBuffer().setContents(previewContent);
                        newWcs.add(wc);                        
                    }
                }
            }
        }
        
        List<BindingWrapper> newBindings = getNewBindings(newWcs, project);
        boolean changed = typeHierarchy.updateWrappers(newBindings);
        
        for (ICompilationUnit unit : newWcs) {
            unit.discardWorkingCopy();
        }
        return changed;
    }

    private static List<BindingWrapper> getNewBindings(List<ICompilationUnit> newWcs, IJavaProject project) throws JavaModelException {
        final Visitor vis = new Visitor();

        ASTParser parser = ASTParser.newParser(AST.JLS3);
        parser.setResolveBindings(true);
        parser.setStatementsRecovery(true);
        parser.setProject(project);
        
        parser.createASTs(newWcs.toArray(new ICompilationUnit[newWcs.size()]), new String[0], 
                new ASTRequestor() {
                    @Override
                    public void acceptAST(ICompilationUnit source, CompilationUnit ast) {
                        ast.accept(vis);
                    }
            
        }, null);
        
        return vis.getBindingWrapper();
    }

    private static class ChangeableInteger {
        int val;
    }
    
    private static class Visitor extends ASTVisitor {
        List<BindingWrapper> bindingWrapper = new ArrayList<BindingWrapper>();
        Stack<ChangeableInteger> methodPosStack = new Stack<ChangeableInteger>();  

        /* (non-Javadoc)
         * @see org.eclipse.jdt.core.dom.ASTVisitor#endVisit(org.eclipse.jdt.core.dom.AnonymousClassDeclaration)
         */
        @Override
        public void endVisit(AnonymousClassDeclaration node) {
            methodPosStack.pop();
            super.endVisit(node);
        }

        /* (non-Javadoc)
         * @see org.eclipse.jdt.core.dom.ASTVisitor#endVisit(org.eclipse.jdt.core.dom.TypeDeclaration)
         */
        @Override
        public void endVisit(TypeDeclaration node) {
            methodPosStack.pop();
            super.endVisit(node);
        }

        /* (non-Javadoc)
         * @see org.eclipse.jdt.core.dom.ASTVisitor#visit(org.eclipse.jdt.core.dom.AnonymousClassDeclaration)
         */
        @Override
        public boolean visit(AnonymousClassDeclaration node) {
            methodPosStack.push(new ChangeableInteger());
            return super.visit(node);
        }

        /* (non-Javadoc)
         * @see org.eclipse.jdt.core.dom.ASTVisitor#visit(org.eclipse.jdt.core.dom.TypeDeclaration)
         */
        @Override
        public boolean visit(TypeDeclaration node) {
            methodPosStack.push(new ChangeableInteger());
            return super.visit(node);
        }

        
        /* (non-Javadoc)
         * @see org.eclipse.jdt.core.dom.ASTVisitor#visit(org.eclipse.jdt.core.dom.EnumDeclaration)
         */
        @Override
        public boolean visit(EnumDeclaration node) {
            methodPosStack.push(new ChangeableInteger());
            return super.visit(node);
        }

        /* (non-Javadoc)
         * @see org.eclipse.jdt.core.dom.ASTVisitor#endVisit(org.eclipse.jdt.core.dom.EnumDeclaration)
         */
        @Override
        public void endVisit(EnumDeclaration node) {
            methodPosStack.pop();
            super.endVisit(node);
        }

        /* (non-Javadoc)
         * @see org.eclipse.jdt.core.dom.ASTVisitor#visit(org.eclipse.jdt.core.dom.MethodDeclaration)
         */
        @Override
        public boolean visit(MethodDeclaration node) {
            IMethodBinding binding = node.resolveBinding();
            
            if (binding != null && !binding.isSynthetic() && !binding.isDefaultConstructor()) {            
                ChangeableInteger pos = methodPosStack.peek();
                bindingWrapper.add(new BindingWrapper(binding, pos.val++));                               
            }
            return super.visit(node);
        }
        
        List<BindingWrapper> getBindingWrapper() throws JavaModelException {
            return bindingWrapper;
        }
        
    }


}
