/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.runtime.Assert;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.variables.DeclaringTypeVariable;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;

/**
 * This class represents an access set for a type variable.
 * <p>
 * In addition to explicitly holding all public instance methods transitively
 * invoked on the element(s) represented by the constraint variable(s), this class also
 * holds the information whether the type of the constraint variable(s) is unmodifiable 
 * (for example due to field access, external types etc.) and if yes, a 
 * description that can be presented to the user.  
 * </p>
 * <p>
 * One AccessSet is used for more than one constraint variable if the variables are inside
 * a cycle or for variables connected with equality constraints.
 * </p>
 * <p>
 * For more information about the calculation of the sets, please
 * see {@link AccessSetCalculation}.
 * </p>
 * @author Hannes Kegel
 *
 */
public class AccessSet {

    private BindingSet members = new BindingSet();
    private boolean unmodifiable = false;
    private Set<String> unmodifiableDescriptions;
    
    private static Set<String> objectSignatures;
    
    private Set<ConstraintVariable2> involvedCvs = new HashSet<ConstraintVariable2>();
    
    /**
     * The constructor used for non-initial access sets.
     * 
     * @param cv the constraint variable this set is created for
     */
    public AccessSet(ConstraintVariable2 cv) {
        involvedCvs.add(cv);
    }
        
    /**
     * Creates an empty set.
     * 
     */
    public AccessSet() {
    }

    /**
     * The constructor for the initial unmodifiable access sets.
     * 
     * @param unmodifiableDescription a description why it is unmodifiable
     * @param cv the constraint variable this set is created for
     */
    public AccessSet(String unmodifiableDescription, ConstraintVariable2 cv) {
        this(cv);
        unmodifiable = true;
        unmodifiableDescriptions = Collections.singleton(unmodifiableDescription);
    }

    /**
     * The constructor for the initial access sets for {@link DeclaringTypeVariable} instances.
     * 
     * @param method the method invoked (must not be null)
     * @param cv the constraint variable this set is created for
     */
    public AccessSet(BindingWrapper method, ConstraintVariable2 cv) {
        this(cv);
        Assert.isTrue(method != null);
        if (!isObjectMethod(method.getMethodBinding())) {
            members.add(method);
        }
    }
    
    /**
     * Merges the argument with this set. All involved constraint variables
     * of the argument will be assigned this access set.
     * 
     * @param otherAccessSet the set to be merged
     */
    public void mergeSet(AccessSet otherAccessSet) {
        if (otherAccessSet == this) {
            return;
        }
        addSet(otherAccessSet);
        involvedCvs.addAll(otherAccessSet.involvedCvs);
        for (ConstraintVariable2 otherCvs : otherAccessSet.involvedCvs) {
            InferTypeConstraintsModel.setAccessSet(otherCvs, this);
        }
    }
    

    /**
     * Add the argument to this set.
     * 
     * @param rightSet the set to be added
     */
    public void addSet(AccessSet rightSet) {
        members.addAll(rightSet.members);
        if (!unmodifiable) {
            unmodifiable = rightSet.unmodifiable;
        }
        if (rightSet.unmodifiableDescriptions != null) {
            if (unmodifiableDescriptions == null) {
                unmodifiableDescriptions = new HashSet<String>();
            } 
            unmodifiableDescriptions.addAll(rightSet.unmodifiableDescriptions);
        }
    }
       
    /**
     * Returns the description why this set is unmodifiable that can be displayed
     * in a message box.
     * 
     * @return the description or null
     */
    public String getShortUnmodifiableDescription() {
        if (unmodifiableDescriptions == null) {
            return null;
        }
        int count = 0;
        String ret = null;
        for (String descr : unmodifiableDescriptions) {
            if (ret == null) {
                ret = descr;
            } else {
                ret += "\n" + descr;
            }
            
            count++;
            if (count == 5) {
                break;
            }
        }
        return ret;
    }
    

    /**
     * Checks if this set is unmodifiable.
     * 
     * @return true if unmodifiable
     */
    public boolean isUnmodfiable() {
        return unmodifiable;
    }

    /**
     * Returns all public instance methods that are in this set.
     * 
     * @return a BindingSet with all public instance methods
     */
    public BindingSet getMembers() {
        return members;
    }
    
    /**
     * Checks if the set contains no methods.
     * 
     * @return true if there are no methods in this set
     */
    public boolean isEmpty() {
        return members.isEmpty();
    }

    /**
     * Returns the number of public instance methods in this set.
     * 
     * @return int
     */
    public int size() {
        return members.size();
    }
    

    /* (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        StringBuffer buf = new StringBuffer();
        buf.append("length: " + members.size() + "\n");
        for (BindingWrapper member : members) {
            buf.append("m: " + member + "\n");
        }
        buf.append("unmodfiable: " + unmodifiable + "\n");
        return buf.toString();
    }

    public static boolean isObjectMethod(IMethodBinding method) {
        //TODO: marker: this is different in IT2, so disable it for comparisons when IT3 runs over IT2 results
        if (objectSignatures == null) {
            objectSignatures = new HashSet<String>();
            objectSignatures.add(".wait(J)V");
            objectSignatures.add(".wait()V");
            objectSignatures.add(".wait(JI)V");
            objectSignatures.add(".notify()V");
            objectSignatures.add(".notifyAll()V");
            objectSignatures.add(".toString()Ljava/lang/String;");
            objectSignatures.add(".hashCode()I");
            objectSignatures.add(".equals(Ljava/lang/Object;)Z");
            objectSignatures.add(".getClass()Ljava/lang/Class;");
            objectSignatures.add(".getClass()Ljava/lang/Class<*>;");
        }
        return objectSignatures.contains(getMethodSignature(method.getKey()));
    }

    private static String getMethodSignature(String bindingKey) {
        return bindingKey.substring(bindingKey.indexOf(';') + 1);
    }
    
}
