/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc;

import java.util.HashSet;
import java.util.Set;

import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.HierarchyType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.WildcardType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ITypeConstraint2;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.InferTypePlugin;
import org.intoJ.inferType3.InferTypeRefactoring;
import org.intoJ.inferType3.internal.constraintModel.EqualityTypeConstraint;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.constraintModel.util.ConstraintTraversal;
import org.intoJ.inferType3.internal.constraintModel.variables.DeclaringTypeVariable;
import org.intoJ.inferType3.internal.constraintModel.variables.ElementVariable;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.BindingWrapper;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.ExistingTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.NewTCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType;
import org.intoJ.inferType3.internal.typeCalc.hierarchy.TypeHierarchy;

/**
 * After the access sets have been calculated, this class determines which existing or new
 * types have to be used so that all constraints are satisfied.
 * <p>
 * In a depth-first pre-order run the spanning tree of the sub-component of the
 * constraint graph with the selected element as root is visited and starting from the initially
 * determined type, new types are calculated that meet the subtyping requirements. For equality constraints
 * the new type is simply transfered to the other variable of the constraint.
 * </p>
 * <p>
 * The handling of a single subtype constraint is very similar to that of Infer Type 2.
 * </p>
 * 
 * @author Hannes Kegel
 *
 */
public class TypeCalculation {
    private InferTypeConstraintsModel model;
    
    private Set<ITypeConstraint2> processSet = new HashSet<ITypeConstraint2>();
    
    private TypeHierarchy typeHierarchy;
    private boolean noAbstractClasses;
    private boolean changeTypeArguments;
    
    /**
     * The single constructor.
     * 
     * @param model the model, access sets have to be calculated already 
     * @param typeHierarchy the hierarchy of the selected type. Contains only existing types
     * at this time.
     * @param noAbstractClasses whether no abstract classes should be generated for case 3
     * @param changeTypeArguments whether type arguments are allowed to be changed if necessary
     * @throws JavaModelException 
     */
    public TypeCalculation(InferTypeConstraintsModel model, TypeHierarchy typeHierarchy, boolean noAbstractClasses, boolean changeTypeArguments) throws JavaModelException {
        this.model = model;
        this.typeHierarchy = typeHierarchy;
        this.noAbstractClasses = noAbstractClasses;
        this.changeTypeArguments = changeTypeArguments;
    }
    
    /**
     * Determine the new types as outlined in the description of this class.
     * 
     * @param newTypeName the name for the initial new type
     * @param checkedMembers the additional members that should be included in the initial type
     * @param checkedInterfaces the interfaces that the new type should be a subtype of
     * @param transitive if true, all connected constraints will be traversed and more than one
     * new type could be created.
     * @param status the refactoring status used to record errors
     * @throws JavaModelException
     */
    public void determineTypes(String newTypeName, Iterable<BindingWrapper> checkedMembers, 
            Iterable<ExistingTCType> checkedInterfaces, boolean transitive, RefactoringStatus status) throws JavaModelException 
            {

        ConstraintVariable2 selectedVar = model.getSelectedConstraintVariable();        

        if (checkedMembers != null) {
            AccessSet accessSet = InferTypeConstraintsModel.getAccessSet(selectedVar);
            assert(accessSet != null);

            for (BindingWrapper wrapper : checkedMembers) {
                accessSet.getMembers().add(wrapper);            
            }
        }
        
        TCType initialNewType = typeHierarchy.getInitialMinimalType(selectedVar, newTypeName, status);

        if (checkedInterfaces != null && initialNewType instanceof NewTCType) {
            for (ExistingTCType interf : checkedInterfaces) {
                if (!initialNewType.equals(interf) && !initialNewType.getAllSupertypes().contains(interf)) {
                    initialNewType.addInterface(interf, status);
                }
            }
        }
        
        TCType oldType = typeHierarchy.getType(selectedVar.getType());
        
        if (initialNewType.equals(oldType)) {
            status.addInfo("The type is already minimal. No changes will be performed.");
            return;
        }

        InferTypeConstraintsModel.setNewType(selectedVar, initialNewType);
        
        if (transitive) {
            ConstraintTraversal.visitSingleStep(processSet, selectedVar, new ConstraintVisitor(status));
        }
        
        //LogUtil.debugOut(this, "TYPES DETERMINED");
    }
        
    /**
     * Returns the hierarchy of the existing and new types of the selection.
     * 
     * @return TypeHierarchy
     */
    public TypeHierarchy getTypeHierarchy() {
        return typeHierarchy;
    }

    private boolean handleSubtypeConstraint(ConstraintVariable2 baseVar, ConstraintVariable2 otherVar, ITypeConstraint2 constraint, RefactoringStatus status) throws JavaModelException {
        TCType newOtherType = null;
        
        TCType oldBaseType = typeHierarchy.getType(baseVar.getType());
        if (oldBaseType == null) {
            //LogUtil.debugOut2(this, "no bounds2 for " + baseVar + ", " + baseVar.getClass() + ", " + constraint);
            return true;
        }

        TCType newBaseType = InferTypeConstraintsModel.getNewType(baseVar);
        if (newBaseType == null) {
            newBaseType = oldBaseType;
        }

        
        TCType prevNewOtherType = InferTypeConstraintsModel.getNewType(otherVar);
        TCType oldOtherType = typeHierarchy.getType(otherVar.getType());
        if (oldOtherType == null) {
            //LogUtil.debugOut2(this, "no bounds1 for " + otherVar + ", " + otherVar.getClass() + ", " + constraint);
            return true;            
        }
        if (oldOtherType.isObject() || newBaseType.equals(oldOtherType)) {
            // do nothing
        } else if (prevNewOtherType != null) {
            if (!newBaseType.equals(prevNewOtherType) && !newBaseType.getAllSupertypes().contains(prevNewOtherType)) {
              //TODO: hannes: does that happen?
                if (newBaseType.canBeMovedInHierarchy()) {
                    prevNewOtherType.setArgumentTType(otherVar.getType());
                    typeHierarchy.addSupertype(newBaseType, prevNewOtherType, status);
                } else {
                    // case 4 behavior
                    newOtherType = typeHierarchy.getType(otherVar, newBaseType, status);
                    newOtherType.setArgumentTType(otherVar.getType());
                    typeHierarchy.addSupertype(oldOtherType, newOtherType, status);
                }
                
            }
        } else if (oldBaseType.equals(oldOtherType)) {
            //////////
            // case 1
            //
            // A = B
            newOtherType = newBaseType;
        } else {

            
            BindingSet allOtherMethodBindings = oldOtherType.getAllMethods();
            
            /*
            LogUtil.debugOut(this, oldBaseType.getTypeName());
            LogUtil.debugOut(this, oldOtherType.getTypeName());
            LogUtil.debugOut(this, "contains: " + newBaseType.containsAll(oldOtherType));
            LogUtil.debugOut(this, "if: " + !oldOtherType.isInterface() + ", if subtype: " + !typeHierarchy.hasInterfaceSubtype(newBaseType) + ", direct: " + oldBaseType.hasDirectSupertype(oldOtherType));
            for (BindingWrapper wrapper : allOtherMethodBindings) {
                LogUtil.debugOut(this, wrapper.getMethodBinding().getKey());                
            }
            LogUtil.debugOut(this, "---");
            */

            //////////
            // case 2
            //
            // B < I  iff  I true subset of B
            //
            if (newBaseType.canBeMovedInHierarchy() && 
                    (allOtherMethodBindings.size() != newBaseType.getAllMethods().size()
                    || oldOtherType.hasOtherDeclaredMembers())  
                    && newBaseType.isContainedIn(allOtherMethodBindings)
                    && newBaseType.isModifiable()) 
            {
                typeHierarchy.addSupertype(oldOtherType, newBaseType, status);
                newBaseType.setArgumentTType(otherVar.getType());
                newOtherType = newBaseType;
            }
            //////////
            // case 3
            //
            // I <= B  iff B subset of I
            //
            else if (newBaseType.isModifiable() && oldBaseType.isModifiable() && 
                    !(newBaseType.isGenericType() || oldBaseType.isGenericType()) &&
                    newBaseType.containsAll(oldOtherType) && ((oldOtherType.isInterface()
                            && !oldOtherType.getAllSupertypes().contains(newBaseType))
                    || (!noAbstractClasses && !typeHierarchy.hasInterfaceSubtype(newBaseType) &&
                            newBaseType.containsAll(oldBaseType.getSuperType())))) 
            {
                //LogUtil.debugOut(this, "case 3: " + newBaseType.getTypeName() + ", " + oldBaseType.getTypeName() + ", " + oldOtherType.getTypeName() + ":: " + constraint + ", type: " + constraint.getClass().getName() + ", otherVar: " + otherVar.getClass().getName());
                if (!newBaseType.getAllSupertypes().contains(oldOtherType)) {
                    if (oldOtherType.isInterface()) {
                        newBaseType.addInterface(oldOtherType, status);
                        typeHierarchy.addSupertype(oldBaseType, newBaseType, status);
                    } else if (newBaseType instanceof ExistingTCType) {
                        if (newBaseType.isContainedIn(allOtherMethodBindings)) {
                            // case 2 behavior for existing classes
                            typeHierarchy.addSupertype(oldOtherType, newBaseType, status);
                            newBaseType.setArgumentTType(otherVar.getType());
                            newOtherType = newBaseType;
                        } else {
                            newOtherType = typeHierarchy.getType(otherVar, newBaseType, status);
                            newOtherType.setArgumentTType(otherVar.getType());
                            typeHierarchy.addSupertype(oldOtherType, newOtherType, status);
                        }
                    } else {
                        newBaseType.setSuperType(oldBaseType.getSuperType(), status);
                        typeHierarchy.addSupertype(oldBaseType, newBaseType, status);
                    }
                }
                //LogUtil.debugOut(this, "case 3 done");
            }

            //////////
            // case 4
            //
            // not B < I  and  not I <= B
            //
            else
            {
                newOtherType = typeHierarchy.getType(otherVar, newBaseType, status);
                newOtherType.setArgumentTType(otherVar.getType());
                typeHierarchy.addSupertype(oldOtherType, newOtherType, status);
            }
        }

        //LogUtil.debugOut(this, "new other type: " + newOtherType + ", " + otherVar.getType().getClass());
        if (!status.hasFatalError() && newOtherType != null && !newOtherType.equals(oldOtherType) && 
                (otherVar.getType() instanceof HierarchyType || otherVar.getType() instanceof WildcardType || otherVar.getType() == null)) {
            if (!oldOtherType.isModifiable()) {
                status.addFatalError("Unmodifiable type " + oldOtherType.getTypeName() + " would have to be changed.");
            } else {
                if (!checkTypeArgChanges(otherVar, newOtherType, status)) {
                    return false;
                }
                
                InferTypeConstraintsModel.setNewType(otherVar, newOtherType);
            }
        }
        return !status.hasFatalError();
    }

    private boolean checkTypeArgChanges(ConstraintVariable2 otherVar, TCType newOtherType, RefactoringStatus status) {
        if (!changeTypeArguments && otherVar instanceof ElementVariable
                && !otherVar.getType().equals(newOtherType)) 
        {
            status.addEntry(RefactoringStatus.FATAL, 
                    "Type arguments would have to be changed", 
                    null, InferTypePlugin.PLUGIN_ID, 
                    InferTypeRefactoring.ERROR_CODE_TYPE_ARGUMENT_MODIFICATION_NECESSARY);
            status.addInfo("You can rerun Infer Type with the 'change type arguments if necessary' option checked");
            return false;
        }
        return true;
    }

    private class ConstraintVisitor implements ConstraintTraversal.Visitor {
        private final RefactoringStatus status;

        public ConstraintVisitor(RefactoringStatus status) {
            this.status = status;
        }

        public boolean visit(ConstraintVariable2 baseVar, ConstraintVariable2 otherVar, ITypeConstraint2 constraint) throws JavaModelException {
            if (constraint instanceof EqualityTypeConstraint) {
                TCType newType = InferTypeConstraintsModel.getNewType(baseVar);
                if (!checkTypeArgChanges(otherVar, newType, status)) {
                    return false;
                }
                
                //LogUtil.debugOut(this, "following: " + constraint);
                InferTypeConstraintsModel.setNewType(otherVar, newType);
            } else if (!(otherVar instanceof DeclaringTypeVariable)) { 
                //LogUtil.debugOut(this, "following: " + constraint);
                if (!handleSubtypeConstraint(baseVar, otherVar, constraint, status)) {
                    return false;
                }
            }
            return ConstraintTraversal.visitSingleStep(processSet, otherVar, this);
        }        
    }
    

}
