/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc;

import java.util.HashSet;
import java.util.Set;

import org.eclipse.core.internal.resources.Resource;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.HierarchyType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints2.ConstraintVariable2;
import org.intoJ.inferType3.internal.constraintModel.InferTypeConstraintsModel;
import org.intoJ.inferType3.internal.manipulation.GenerationUtil;

/**
 * <p>This is the (one and only) InferType type name generator.</p>
 * 
 * <p>The generator generates names for new types for a certain constraint variable. 
 * It is guaranteed that the given type name is a valid and unambiguous one. It looks like this:</p>
 * 
 * <pre>
 * SimpleName(declaredType) _ SimpleName(DE)    _of_SimpleName(method)  _ x  IT
 *
 *                                              | only                |   x: unambiguous  
 *                                              | for                 |      number
 *                                              | parameters          |
 * </pre>
 * 
 * <p>Note that the first element (the declared type) is ALWAYS the original type of the DE, not
 * an InferType3-induced type. For example, if the DE already has a type "SomeClass_field1_1IT",
 * only the first part ("SomeClass") will be re-used for the new type name.</p>
 * 
 * <p>Note also that each name is unambiguous for the complete project (regardless of the package).</p>
 * 
 * @author Philip Mayer
 * @author Hannes Kegel
 */
public class TypeNameGenerator {

    private Set<String> usedNames = new HashSet<String>();
    private IJavaProject javaProject;

    /**
     * The single constructor.
     * 
	 * @param javaProject the current IJavaProject
	 */
	public TypeNameGenerator(IJavaProject javaProject) {
        this.javaProject = javaProject;
	}

	/**
     * Creates a new type name for the type used for the given {@link ConstraintVariable2}.
	 * @param selectedVar the variable that should be linked to a new type
	 * @param previousType the hierarchy type derived from the previous type of the constraint variable
     * @param packageName the package name at which the new type sould be added.
     * <code>null</code> if it should be derived automatically 
	 * @return the canoncial name
	 */
	public String getName(ConstraintVariable2 selectedVar,
			HierarchyType previousType, String packageName) {
        
		String postfix = "IT";
		String name;

        //LogUtil.debugOut(this, "getting name for " + selectedVar + ", " + selectedVar.getClass() + ", " + InferTypeConstraintsModel.getSimpleName(selectedVar));

		// Add type prefix

		String oldName = previousType.getName();
        int parameterStart = oldName.indexOf('<');
        if (parameterStart > -1) {
            oldName = oldName.substring(0, parameterStart);
        }
        
        if (oldName.endsWith(postfix)) {
			// is IT type! get first 'element' before _ (is original type name)
			name = oldName.substring(0, oldName.indexOf('_'));
		} else {
			// is original type, get name
			name = transformName(oldName);
		}

		// Add simple name of de
		String transformedName = transformName(InferTypeConstraintsModel.getSimpleName(selectedVar));
        name += transformedName.equals("") ? "" : "_" + transformedName;

        String methodName = InferTypeConstraintsModel.getMethodName(selectedVar);
		if (methodName != null) {
            name += "_of_" + InferTypeConstraintsModel.getMethodName(selectedVar);
        }

		// append count suffix and render the name unambigous. 
		int counter = 1;
		if (packageName == null) {
			packageName = determinePackage(previousType);
		}
		name = name + "_" + counter + postfix;

		while (!isValid(packageName, name)) {

			name = name.substring(0, name.lastIndexOf("_"));
			name += "_" + counter + postfix;
			counter++;
		}
        
		usedNames.add(name);

        //LogUtil.debugOut2(this, "simple name " + InferTypeConstraintsModel.getSimpleName(selectedVar));
        //LogUtil.debugOut2(this, "used name is " + packageName + (packageName.equals("") ? "" : ".") + name);
		return packageName + (packageName.equals("") ? "" : ".") + name;
	}

	/**
	 * This method transforms the declaration elements simple name to a valid
	 * type name by removing characters $, (, ), _, and ','.
	 * 
	 * The first character is uppercased.
	 *  
	 * @param simpleName the simple name of a declaration element
	 * @return a valid name for a type
	 */
	private static String transformName(String simpleName) {
        if (simpleName == null) {
            return "";
        }
		StringBuffer name = new StringBuffer(simpleName.replaceAll("[$()_,]", ""));

		if (name.length() > 1) {
            return Character.toUpperCase(name.charAt(0)) + name.substring(1);
        } else {
            return name.toString();
        }

	}
    
    private boolean isValid(String thePackage, String theSimpleName) {
        try {
            if (usedNames.contains(theSimpleName)) {
                return false;
            }
            
            String fqName = thePackage + "." + theSimpleName;
            if (javaProject.findType(fqName) != null) {
                return false;
            }
            
            ICompilationUnit compilationUnit = GenerationUtil.createCompilationUnit(fqName, javaProject);
            Resource resource = (Resource)compilationUnit.getResource();
            resource.checkDoesNotExist(resource.getFlags(resource.getResourceInfo(false, false)), false);
            return true;
        } catch (CoreException e) {
            return false;
        }
    }
    
    private static String determinePackage(HierarchyType type) {
        return type.getJavaElementType().getPackageFragment().getElementName();
    }


}
