/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc.hierarchy;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.IVariableBinding;
import org.eclipse.jdt.core.dom.Modifier;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.ltk.core.refactoring.RefactoringStatus;
import org.intoJ.inferType3.InferTypePlugin;
import org.intoJ.inferType3.internal.typeCalc.BindingSet;

/**
 * A TCType in the hierarchy of the selected type that represent a 
 * previously existing type. The supertype or the declared interfaces of this
 * type may change.
 * 
 * @author Hannes Kegel
 *
 */
public class ExistingTCType extends TCType {
    
    private final BindingSet declaredMethods = new BindingSet(); 
    private BindingSet allMethods; 
    
    private TCType oldSuperType;
    private final Set<TCType> oldInterfaces = new HashSet<TCType>();
    
    private Set<String> removedInterfaces = new HashSet<String>();

    private boolean hasOtherAccessibleMembers = false;
    private boolean hasOtherDeclaredMembers = false;
    
    private Set<String> otherAccessibleMemberDescriptions = new HashSet<String>();
    private Set<String> otherDeclaredMemberDescriptions = new HashSet<String>();
    
    private boolean object;
    private boolean modifiable;

    /**
     * The single constructor.
     * 
     * @param baseType the type binding of the existing type
     * @param superType the previously created super type
     * @param interfaces the previously created interfaces
     * @throws JavaModelException
     */
    public ExistingTCType(ITypeBinding baseType, TCType superType, Set<TCType> interfaces, boolean allMethods) throws JavaModelException {
        super(baseType.getErasure());
        
        this.superType = superType;
        this.interfaces = interfaces;
        oldSuperType = superType;
        oldInterfaces.addAll(interfaces);
        
        object = baseType.getQualifiedName().equals("java.lang.Object");

        for (IMethodBinding methodBinding : baseType.getDeclaredMethods()) {
            if (!methodBinding.isConstructor() && (allMethods  || !Modifier.isStatic(methodBinding.getModifiers()))) {
                if (allMethods || Modifier.isPublic(methodBinding.getModifiers())) {
                    declaredMethods.add(new BindingWrapper(methodBinding));                    
                } else if (Modifier.isProtected(methodBinding.getModifiers())) {
                    hasOtherAccessibleMembers = true;
                    hasOtherDeclaredMembers = true;
                    checkModifiers(methodBinding.getModifiers(), "methods");
                } else {
                    hasOtherDeclaredMembers = true;
                    checkModifiers(methodBinding.getModifiers(), "methods");
                }
            }
        }
        
        for (IVariableBinding variableBinding : baseType.getDeclaredFields()) {
            if (Modifier.isPublic(variableBinding.getModifiers()) ||
                     (Modifier.isProtected(variableBinding.getModifiers()))) 
            {
                hasOtherAccessibleMembers = true;
                hasOtherDeclaredMembers = true;                
            } else {
                hasOtherDeclaredMembers = true;                
            }
            checkModifiers(variableBinding.getModifiers(), "fields");
        }
        
        modifiable = baseType.isFromSource() && !isObject();
    }
    
    private void checkModifiers(int modifiers, String memberType) {
        if (Modifier.isProtected(modifiers)) {
            otherAccessibleMemberDescriptions.add("protected " + memberType);
        } else if (Modifier.isPublic(modifiers)) {
            otherAccessibleMemberDescriptions.add("public " + memberType);
        } else if (Modifier.isPrivate(modifiers)) {
            otherDeclaredMemberDescriptions.add("private " + memberType);
        } else {
            otherDeclaredMemberDescriptions.add("package-private " + memberType);
        }
    }

    /**
     * Returns the new supertype if there was a change.
     * 
     * @return the new type or null
     */
    public TCType getModifiedSuperType() {
        if (superType != oldSuperType) {
            return superType;
        }
        return null;
    }

    /**
     * Checks if the hierarchy declaration of this type has been changed.
     *  
     * @return true if there have been changes
     */
    public boolean isModified() {
        return superType != oldSuperType || !oldInterfaces.equals(interfaces);
    }

    /**
     * Gets all declared non-static, public methods.
     * 
     * @return BindingSet
     */
    public BindingSet getDeclaredMethods() {
        return declaredMethods;
    }
    
    /**
     * Get the corresponding IType.
     * 
     * @return the IType of the base type binding
     */
    public IType getIType() {
        return (IType)baseTypeBinding.getJavaElement();
    }

    /**
     * Retrieves all new interfaces that the base type should implement.
     * 
     * @return Collection<TCType>
     */
    public Collection<TCType> getNewInterfaces() {
        List<TCType> list = new ArrayList<TCType>();
        for (TCType currentIf : interfaces) {
            if (!oldInterfaces.contains(currentIf)) {
                list.add(currentIf);
            }
        }
        return list;
    }

    /**
     * Retrieves all interfaces that should be removed from the base type.
     * 
     * @return a set of type bindings
     */
    public Set<String> getRemovedInterfaces() {
        return removedInterfaces;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#isInterface()
     */
    @Override
    public boolean isInterface() {
        return baseTypeBinding.isInterface();
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#getTypeName()
     */
    @Override
    public String getTypeName() {
        return baseTypeBinding.getQualifiedName();
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#getAllMethods()
     */
    @Override
    public BindingSet getAllMethods() {
        if (allMethods == null) {
            allMethods = new BindingSet();
            allMethods.addAll(declaredMethods);

            //this could be used instead of addAll to completely ignore object methods
            //currently not used for compatibility reasons with IT2.
            /*
            for (BindingWrapper bindingWrapper : declaredMethods) {
                if (!AccessSet.isObjectMethod(bindingWrapper.getMethodBinding())) {
                    allMethods.add(bindingWrapper);
                }
            }
            */
            
            if (superType != null && !superType.isObject()) {
                allMethods.addAll(superType.getAllMethods());
            }
            for (TCType usedType : interfaces) {
                allMethods.addAll(usedType.getAllMethods());                
            }
        }
        return allMethods;
    }
    
    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#removeInterfaceInt(org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType)
     */
    @Override
    protected void removeInterfaceInt(TCType interf) {
        if (interf instanceof ExistingTCType) {
            removedInterfaces.add(((ExistingTCType)interf).baseTypeBinding.getKey());
        }
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#hasOtherAccessibleMembers()
     */
    @Override
    public boolean hasOtherAccessibleMembers() {        
        return !object && (hasOtherAccessibleMembers || super.hasOtherAccessibleMembers());
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#isObject()
     */
    @Override
    public boolean isObject() {
        return object;
    }
    
    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#setSuperType(org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType, org.eclipse.ltk.core.refactoring.RefactoringStatus)
     */
    @Override
    public void setSuperType(TCType superType, RefactoringStatus status) throws JavaModelException {
        if (isInterface()) {
            status.addFatalError("tried to set superclass for interface " + getTypeName() + " to " + superType.getTypeName());
            return;
        }
        super.setSuperType(superType, status);
        if (superType instanceof NewTCType) {
            superType.setBaseTypeBinding(baseTypeBinding);
        }
    }
    
    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#addInterface(org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType, org.eclipse.ltk.core.refactoring.RefactoringStatus)
     */
    @Override
    public boolean addInterface(TCType newInterface, RefactoringStatus status) throws JavaModelException {
        if (super.addInterface(newInterface, status)) {
            if (newInterface instanceof NewTCType) {
                newInterface.setBaseTypeBinding(baseTypeBinding);
            }
            
            return true;
        }
        return false;        
    }
    
    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#canBeMovedInHierarchy()
     */
    @Override
    public boolean canBeMovedInHierarchy() {
        return !isGenericType();
    }
    
    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#setArgumentTType(org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType)
     */
    @Override
    public void setArgumentTType(TType argumentTType) {
        super.setArgumentTType(findExistingTType(baseTypeBinding.getKey(), argumentTType));
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#isModifiable()
     */
    @Override
    public boolean isModifiable() {
        return modifiable;
    }

    /* (non-Javadoc)
     * @see org.intoJ.inferType3.internal.typeCalc.hierarchy.TCType#hasOtherDeclaredMembers()
     */
    @Override
    public boolean hasOtherDeclaredMembers() {
        return hasOtherDeclaredMembers;
    }
    
    private static TType findExistingTType(String wantedBindingKey, TType argumentTType) {
        if (argumentTType.getErasure().getBindingKey().equals(wantedBindingKey)) {
            return argumentTType;
        }
        TType superBinding =  argumentTType.getSuperclass();
        if (superBinding != null) {
            TType ret = findExistingTType(wantedBindingKey, superBinding);
            if (ret != null) {
                return ret;
            }
        }
        
        for (TType interf : argumentTType.getInterfaces()) {
            TType ret = findExistingTType(wantedBindingKey, interf);
            if (ret != null) {
                return ret;
            }            
        }
                
        return null;
    }
    
    private Set<String> getOtherMemberDescriptions(boolean all) {
        Set<String> ret = new HashSet<String>();
        if (!isObject()) {
            if (otherAccessibleMemberDescriptions != null) {
                ret.addAll(otherAccessibleMemberDescriptions);
            }
            if (all && otherDeclaredMemberDescriptions != null) {
                ret.addAll(otherDeclaredMemberDescriptions);
            }
            
            Set<TCType> allSupertypes = getDeclaredSupertypes();
            for (TCType type : allSupertypes) {
                if (type instanceof ExistingTCType) {
                    ret.addAll(((ExistingTCType)type).getOtherMemberDescriptions(false));
                }
            }
        }
        
        return ret;
        
    }
    
    /**
     * Returns a description of other members or null.
     * 
     * @return the description or null
     */
    public String getOtherMemberDescription() {
        Set<String> descriptions = getOtherMemberDescriptions(
                InferTypePlugin.getDefault().getPreferenceStore().getBoolean(InferTypePlugin.PREFNAME_CONSIDER_PRIVATE));
        if (descriptions.isEmpty()) {
            return null;
        }
        
        StringBuilder builder = new StringBuilder("The original type has the following other kinds of members: ");
        for (Iterator<String> it = descriptions.iterator(); it.hasNext();) {
            builder.append(it.next());
            if (it.hasNext()) {
                builder.append(", ");
            }
        }
        return builder.toString();
    }
    
    
}
