/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.typeCalc.hierarchy;

import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.dom.rewrite.ImportRewrite;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.ArrayType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.GenericType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.HierarchyType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.ParameterizedType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.TType;
import org.eclipse.jdt.internal.corext.refactoring.typeconstraints.types.WildcardType;
import org.intoJ.inferType3.internal.util.Cache;
import org.intoJ.inferType3.internal.util.InternalCalculationException;

/**
 * Utility class for getting information about parameterized types. In particular, this class
 * handles the String creation for the type arguments and parameters. 
 *  
 * @author Hannes Kegel
 *
 */
public class ParameterizedTypeUtil {
    /**
     * Returns the type parameters including brackets for the erasure of the given binding.
     * 
     * @param baseType the parameterized type
     * @param importRewrite the import rewrite to get simple names, may be null
     * @return the erasure parameters
     */
    public static String getErasureTypeParameters(ITypeBinding baseType, ImportRewrite importRewrite) {
        return getTypeArguments(baseType.getErasure(), new ImportRewriteProxy(importRewrite));
    }
    
    /**
     * Returns the type arguments or parameters for the given binding.
     * 
     * @param baseType the binding of interest
     * @param importRewrite a proxy to an ImportRewrite
     * @return the type arguments or parameters including brackets or the empty string 
     */
    public static String getTypeArguments(final ITypeBinding baseType, ImportRewriteProxy importRewrite) {
        if (baseType.isGenericType()) {
            return getTypeParameters(baseType.getTypeParameters(), importRewrite);
        } else if (baseType.isParameterizedType()) {
            return getTypeParameters(baseType.getTypeArguments(), importRewrite);            
        }
        return "";
    }

    /**
     * Returns the parameterized binding in the hierarchy of the given type binding
     * whose erasure matches the given binding key.
     * 
     * @param wantedBindingKey the erasure binding key
     * @param subTypeBinding the binding whose hierarchy will be searched
     * @return the parametrized binding
     * @throws JavaModelException
     * @throws InternalCalculationException if the binding could not be found
     */
    public static ITypeBinding getParametrizedTypeBinding(String wantedBindingKey, ITypeBinding subTypeBinding) throws JavaModelException, InternalCalculationException {
        ITypeBinding ret = findInSuperHierarchy(wantedBindingKey, subTypeBinding);
        if (ret != null) {
            return ret;
        }
        
        IType type = (IType)subTypeBinding.getJavaElement();
        ITypeHierarchy subHierarchy = Cache.getHierarchy(null, type, false);
        for (IType subType : subHierarchy.getAllSubtypes(type)) {
            ITypeBinding binding = Cache.getBinding(subType);
            ret = findInSuperHierarchy(wantedBindingKey, binding);
            if (ret != null) {
                return ret;
            }            
        }    
        throw new InternalCalculationException("Could not find parameterized type");
    }

    /**
     * Returns the signature of the type arguments. It will search for the baseType in
     * the hierarchy of the oldType. The baseType is not used directly
     * to ensure the same wildcard usage as previously. 
     * 
     * @param baseType the base type that defines the type arguments used
     * @param oldType the previous type used in the place to be modified 
     * @return the relevant part of the type signature or the empty string
     */
    public static String getTypeArgumentSignature(TType baseType, TType oldType) {
        baseType = stripType(baseType);
        oldType = stripType(oldType);
        
        if (baseType instanceof ParameterizedType) {
            TType parametrizedBaseType = findParametrizedBaseType(baseType.getTypeDeclaration(), oldType);
            if (parametrizedBaseType instanceof ParameterizedType) {
                return getTypeParametersSig(((ParameterizedType)parametrizedBaseType).getTypeArguments());
            }
        } else if (baseType instanceof GenericType) {
            return getTypeParametersSig(((GenericType)baseType).getTypeParameters());
        }
        return "";
    }

    private static TType stripType(TType baseType) {
        if (baseType instanceof WildcardType) {
            baseType = ((WildcardType)baseType).getBound();
        }
        if (baseType instanceof ArrayType) {
            baseType = ((ArrayType)baseType).getElementType();
        }
        return baseType;
    }
    
    private static String getTypeParametersSig(TType[] typeParameters) {
        StringBuilder builder = new StringBuilder();

        builder.append("<");
        for (TType type : typeParameters) {
            builder.append(getTypeSig(type));
        }
        builder.append(">");
        
        return builder.toString();
    }

    private static CharSequence getTypeSig(TType type) {
        StringBuilder builder = new StringBuilder();
        if (type.isWildcardType()) {
            if (type.isUnboundWildcardType()) {
                builder.append(Signature.C_STAR);
            } else if (type.isExtendsWildcardType()) {
                builder.append(Signature.C_EXTENDS);                
            } else {
                builder.append(Signature.C_SUPER);                                
            }
            type = ((WildcardType)type).getBound();
        }
        
        if (type != null) {        
            while (type.isArrayType()) {
                type = ((ArrayType)type).getComponentType();
                builder.append("[");
            }
            if (type.isTypeVariable()) {
                builder.append("T").append(type.getName());
            } else if (type.isHierarchyType()) {
                HierarchyType hierarchyType = (HierarchyType)type;
                builder.append("L").append(hierarchyType.getJavaElementType().getFullyQualifiedName('.'));
                
                if (hierarchyType instanceof ParameterizedType) {
                    builder.append(getTypeParametersSig(((ParameterizedType)hierarchyType).getTypeArguments()));
                } else if (hierarchyType instanceof GenericType) {
                    builder.append(getTypeParametersSig(((GenericType)hierarchyType).getTypeParameters()));
                }
            } else {
                builder.append("L").append(type.getPrettySignature());
            }
            builder.append(";");
        }
        
        return builder;
    }

    // used to preserve wildcards arguments correctly
    private static TType findParametrizedBaseType(TType typeDeclaration, TType oldType) {
        if (oldType.getTypeDeclaration().equals(typeDeclaration)) {
            return oldType;
        }
        if (oldType.getSuperclass() != null) {
            TType ret = findParametrizedBaseType(typeDeclaration, oldType.getSuperclass());
            if (ret != null) {
                return ret;
            }            
        }
        
        for (TType interf : oldType.getInterfaces()) {
            TType ret = findParametrizedBaseType(typeDeclaration, interf);
            if (ret != null) {
                return ret;
            }                        
        }
        return null;
    }

    private static String getTypeParameters(ITypeBinding[] typeParameters, ImportRewriteProxy importRewrite) {
        StringBuilder builder = new StringBuilder();

        builder.append("<");
        for (int i = 0; i < typeParameters.length; i++) {
            builder.append(getPrettySignature(typeParameters[i], importRewrite));
            if (i < typeParameters.length - 1) {
                builder.append(", ");
            }
        }
        builder.append(">");
        
        return builder.toString();
    }

    private static String getPrettySignature(ITypeBinding binding, ImportRewriteProxy importRewrite) {
        return importRewrite.addImport(binding);
    }
    
    private static ITypeBinding findInSuperHierarchy(String wantedBindingKey, ITypeBinding subTypeBinding) throws JavaModelException, InternalCalculationException {
        if (subTypeBinding.getErasure().getKey().equals(wantedBindingKey)) {
            return subTypeBinding;
        }
        ITypeBinding superBinding =  subTypeBinding.getSuperclass();
        if (superBinding != null) {
            ITypeBinding ret = findInSuperHierarchy(wantedBindingKey, superBinding);
            if (ret != null) {
                return ret;
            }
        }
        
        for (ITypeBinding interf : subTypeBinding.getInterfaces()) {
            ITypeBinding ret = findInSuperHierarchy(wantedBindingKey, interf);
            if (ret != null) {
                return ret;
            }            
        }
        return null;
    }
    
    
    /**
     * The import rewrite proxy is used to be able to call all the methods
     * without an {@link ImportRewrite}.
     *  
     * @author Hannes Kegel
     *
     */
    public static class ImportRewriteProxy {
        private ImportRewrite importRewrite;

        /**
         * The single constructor.
         * 
         * @param importRewrite may be null
         */
        public ImportRewriteProxy(ImportRewrite importRewrite) {
            this.importRewrite = importRewrite;
        }

        /**
         * Adds the import and returns a simple name.
         * 
         * @param binding the binding
         * @return a simple name or the qualified name if there is no ImportRewrite
         */
        public String addImport(ITypeBinding binding) {
            if (importRewrite == null) {
                return binding.getQualifiedName();
            } else {
                return importRewrite.addImport(binding);                
            }
        }

        /**
         * Adds the import and returns a simple name.
         * 
         * @param qualifiedName the qualified name
         * @return a simple name or the qualified name if there is no ImportRewrite
         */
        public String addImport(String qualifiedName) {
            if (importRewrite == null) {
                return qualifiedName;
            } else {
                return importRewrite.addImport(qualifiedName);                
            }
        }                       
    }
    
    
    
}
