/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.util;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMember;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.ITypeHierarchy;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.WorkingCopyOwner;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.eclipse.jdt.core.search.IJavaSearchConstants;
import org.eclipse.jdt.core.search.SearchMatch;
import org.eclipse.jdt.core.search.SearchPattern;
import org.eclipse.jdt.internal.corext.refactoring.RefactoringSearchEngine2;
import org.eclipse.jdt.internal.corext.refactoring.SearchResultGroup;
import org.eclipse.jdt.internal.corext.refactoring.util.RefactoringASTParser;
import org.eclipse.jdt.internal.corext.util.SearchUtils;

/**
 * This class bundles all basic operations that are cached for performance reasons.
 * The cache can be persistent for multiple refactoring runs thus speeding up the
 * mulirefac action. The operations are bundled to have a single place to 
 * savely prune the cache.
 * 
 * @author Hannes Kegel
 *
 */
public class Cache {
    private static Map<IType, ITypeHierarchy> superHierarchyCache = new CacheMap<IType, ITypeHierarchy>(600);
    private static Map<IType, ITypeHierarchy> hierarchyCache = new CacheMap<IType, ITypeHierarchy>(400);
    private static Map<IType, ITypeBinding> bindingCache = new CacheMap<IType, ITypeBinding>(70);
    private static Set<IType> unusedHierarchyTypes = new HashSet<IType>();
    private static Map<IJavaElement, Map<IJavaProject, Set<ICompilationUnit>>> referenceCache = new CacheMap<IJavaElement, Map<IJavaProject,Set<ICompilationUnit>>>(300);
    private static Map<IJavaElement, Map<IJavaProject, Collection<SearchResultGroup>>> referenceGroupCache = new CacheMap<IJavaElement, Map<IJavaProject,Collection<SearchResultGroup>>>(80); 
    private static Map compilerOptions;
    
    /**
     * Gets the supertype hierarchy of the given type as in IType.newSupertypeHierarchy.
     * 
     * @param type the type whose supertype hierarchy we want
     * @return the supertype hierarchy
     * @throws JavaModelException
     */
    public static ITypeHierarchy getSuperHierarchy(IType type) throws JavaModelException {
    	if (type == null) {
            return null;
        }
    	
        ITypeHierarchy ret = superHierarchyCache.get(type);
        if (ret == null) {
            ret = type.newSupertypeHierarchy(null);
            superHierarchyCache.put(type, ret);
        }
        return ret;        
    }
    
    /**
     * Gets the hierarchy of the given type as in IType.newTypeHierarchy.
     *
     * @param owner the current owner 
     * @param type the type whose hierarchy we want
     * @param checkUnused whether we should check if the hierarchy was marked unused with markHierarchyAsUnused
     * @return ITypeHierarchy
     * @throws JavaModelException
     */
    public static ITypeHierarchy getHierarchy(WorkingCopyOwner owner, IType type, boolean checkUnused) throws JavaModelException {
        if (type == null || (checkUnused && unusedHierarchyTypes.contains(type))) {
            return null;
        }
        ITypeHierarchy ret = hierarchyCache.get(type);
        if (ret == null) {
            ret = type.newTypeHierarchy(owner, null);
            hierarchyCache.put(type, ret);            
        }
        return ret;        
    }
    
    /**
     * Gets an ITypeBinding for the given IType.
     * 
     * @param type the IType we need an ITypeBinding instead
     * @return the binding
     * @throws InternalCalculationException
     */
    public static ITypeBinding getBinding(IType type) throws InternalCalculationException {
    	if (type == null) {
            return null;
        }

    	ITypeBinding ret = bindingCache.get(type);
        if (ret == null) {
            ret = ConversionUtil.getTypeBinding(type);
            bindingCache.put(type, ret);
        }
        return ret;
    }

    /**
     * Sets a mark that the hierarchy for the given type is unused.
     *  
     * @param type the type
     */
    public static void markHierarchyAsUnused(IType type) {
        unusedHierarchyTypes.add(type);
    }
    
    /**
     * Clear the cache completely
     */
    public static void clearAll() {
        unusedHierarchyTypes.clear();            
        bindingCache.clear();
        superHierarchyCache.clear();
        hierarchyCache.clear();
        referenceCache.clear();
        referenceGroupCache.clear();
        compilerOptions = null;
    }

    /**
     * Prune everything that could have changed. The hierarchy cache
     * must be adressed separately with clearHierarchyCache.
     * 
     * @throws JavaModelException
     */
    public static void changesMade() throws JavaModelException {
        unusedHierarchyTypes.clear();            
        bindingCache.clear();
        superHierarchyCache.clear();
        referenceCache.clear();
        referenceGroupCache.clear();
        // anonymous types change their identifier when the source pos changes, 
        // otherwise hierarchyCache is only cleared by clearHierarchyCache
        for (Iterator<Map.Entry<IType, ITypeHierarchy>> it = hierarchyCache.entrySet().iterator(); it.hasNext();) {
            Map.Entry<IType, ITypeHierarchy> entry = it.next();
            for (IType subType : entry.getValue().getAllTypes()) {
                if (subType.isAnonymous()) {
                    it.remove();
                    break;
                }                
            }
        }
    }

    /**
     * Prune the entry for the given type from the hierarchy cache.
     * 
     * @param type the type to be pruned
     * @throws JavaModelException
     */
    public static void clearHierarchyCache(IType type) throws JavaModelException {
        unusedHierarchyTypes.clear();            
        hierarchyCache.remove(type);        
        ITypeHierarchy superHierarchy = getSuperHierarchy(type);
        for (IType superType : superHierarchy.getAllTypes()) {
            hierarchyCache.remove(superType);                    
        }
    }
    
    /**
     * Find all references of the given element and return the search result groups.
     * 
     * @param javaElement the element to search for
     * @param engine the search engine
     * @return a map from projects to collections of {@link SearchResultGroup}
     * @throws JavaModelException
     */
    @SuppressWarnings("unchecked")
    public static Map<IJavaProject, Collection<SearchResultGroup>> findReferenceGroups(IJavaElement javaElement, RefactoringSearchEngine2 engine) throws JavaModelException {
        Map<IJavaProject, Collection<SearchResultGroup>> groups = referenceGroupCache.get(javaElement);
        if (groups == null) {
            engine.setPattern(SearchPattern.createPattern(javaElement, IJavaSearchConstants.REFERENCES, SearchUtils.GENERICS_AGNOSTIC_MATCH_RULE));
            engine.searchPattern(new NullProgressMonitor());
    
            groups = engine.getAffectedProjects();            
            referenceGroupCache.put(javaElement, groups);
        }
        return Collections.unmodifiableMap(groups);
        
    }

    /**
     * Find all references of the given element and return the compilation units.
     * 
     * @param javaElement the element to search for
     * @param engine the search engine
     * @return a map from projects to sets of {@link ICompilationUnit}
     * @throws JavaModelException
     */
    @SuppressWarnings("unchecked")
    public static Map<IJavaProject, Set<ICompilationUnit>> findReferences(IJavaElement javaElement, RefactoringSearchEngine2 engine) throws JavaModelException {
        Map<IJavaProject, Set<ICompilationUnit>> units = referenceCache.get(javaElement);
        if (units == null) {
            units = new HashMap<IJavaProject, Set<ICompilationUnit>>();
            
            engine.setPattern(SearchPattern.createPattern(javaElement, IJavaSearchConstants.REFERENCES, SearchUtils.GENERICS_AGNOSTIC_MATCH_RULE));
            engine.searchPattern(new NullProgressMonitor());
    
            Map<IJavaProject, Set<SearchResultGroup>> result = engine.getAffectedProjects();
    
            for (Map.Entry<IJavaProject, Set<SearchResultGroup>> entry : result.entrySet()) {
                Set<ICompilationUnit> unitsToBeParsed = units.get(entry.getKey());
                if (unitsToBeParsed == null) {
                    unitsToBeParsed = new HashSet<ICompilationUnit>();
                    units.put(entry.getKey(), unitsToBeParsed);
                }
    
                for (SearchResultGroup group : entry.getValue()) {
                    SearchMatch[] matches = group.getSearchResults();
                    for (SearchMatch element2 : matches) {
                        Object element = element2.getElement();
                        if (element instanceof IMember) {
                            ICompilationUnit current = ((IMember) element).getCompilationUnit();
                            if (current != null) {
                                unitsToBeParsed.add(current);
                            }
                        }
                    }
                }
            }
            
            referenceCache.put(javaElement, units);
        }
        return Collections.unmodifiableMap(units);
    }

    /**
     * Get the current compiler option.
     * 
     * @param project the current project. Only used when the options have to be created.
     * @return the compiler options
     */
    public static Map getCompilerOptions(IJavaProject project) {
        if (compilerOptions == null) {
            compilerOptions = RefactoringASTParser.getCompilerOptions(project);
        }
        return compilerOptions;
    }


}
