/*
 * This file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType3.internal.util;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintWriter;
import java.util.HashSet;
import java.util.Set;
import java.util.StringTokenizer;

/**
 * This writer can be used to create graph files in gml format.
 * <p>
 * The generated files can be read and layouted with a tool like
 * <a href="http://www.yworks.com/en/products_yed_about.htm">yEd</a>.
 * </p>
 * <p>
 * Write all nodes first with {@code writeNode} and the the edges with 
 * {@code writeEdge}. Then call {@code finish}. The connection between
 * nodes and edges is done via {@code System.identityHashCode}.
 * </p>
 * 
 * @author Hannes Kegel
 *
 */
public class GmlWriter {
    private PrintWriter pw;    
    private Set<Integer> ids = new HashSet<Integer>();

    private static int PREFERRED_LINE_SIZE = 40;
    private static int MIN_LINE_SIZE  = 8;
    
    /**
     * The single constructor.
     * 
     * @param file the output file
     * @throws FileNotFoundException
     */
    public GmlWriter(File file) throws FileNotFoundException {
        pw = new PrintWriter(new BufferedOutputStream(new FileOutputStream(file)));
        pw.println("graph");
        pw.println("[");
        pw.println("\tdirected\t1");
    }
    
    /**
     * Writes a node to the file.
     * 
     * @param node any object that is used as a node
     * @param description the description of this node
     * @param color the color of the node formatted like #FFFFFF (may be null)
     */
    public void writeNode(Object node, String description, String color) {
        int identityHashCode = System.identityHashCode(node);
        ids.add(identityHashCode); 

        pw.println("\tnode");
        pw.println("\t[");
        pw.println("\t\tid\t" + identityHashCode);
        if (description != null) {
            pw.println("\t\tlabel\t\"" + wrap(description) + "\"\n");
        }
        if (color != null) {
            pw.println("\t\tgraphics");
            pw.println("\t\t[");
            pw.println("\t\t\tfill\t\"" + color + "\"");
            pw.println("\t\t]");
        }
        
        pw.println("\t\tLabelGraphics");
        pw.println("\t\t[");
        pw.println("\t\t\tfontSize\t12");
        pw.println("\t\t\tfontName\t\"Lucida Sans\"");
        pw.println("\t\t]");
        
        pw.println("\t]");        
    }
    
    /**
     * Writes an edge to the file.
     * 
     * @param source the source node object
     * @param target the target node object
     * @param description the edge description (may be null)
     * @param color the color of the node formatted #FFFFFF (may be null)
     * @param style the style ("dashed", "dotted"), may be null
     * @param targetArrow if the edge should have an arrow pointing to the target
     */
    public void writeEdge(Object source, Object target, String description, String color, String style, boolean targetArrow) {
        pw.println("\tedge");
        pw.println("\t[");
        pw.println("\t\tsource\t" + System.identityHashCode(source));
        pw.println("\t\ttarget\t" + System.identityHashCode(target));
/*
        if (!ids.contains(System.identityHashCode(source))) {
            LogUtil.debugOut2(this, "DOES NOT CONTAIN SOURCE " + source);
        }
        if (!ids.contains(System.identityHashCode(target))) {
            LogUtil.debugOut2(this, "DOES NOT CONTAIN TARGET " + target);
        }
        */
        if (description != null) {
            pw.println("\t\tlabel\t\"" + wrap(description) + "\"\n");
        }
        pw.println("\t\tgraphics");
        pw.println("\t\t[");
        if (style != null) {
            pw.println("\t\t\tstyle\t\"" + style + "\"");
        }
        if (color != null) {
            pw.println("\t\t\tfill\t\"" + color + "\"");
        }
        if (targetArrow) {
            pw.println("\t\t\ttargetArrow\t\"standard\"");
        }
        pw.println("\t\t]");
                
        if (description != null) {
            pw.println("\t\tLabelGraphics");
            pw.println("\t\t[");
            pw.println("\t\t\tfontSize\t12");
            pw.println("\t\t\tfontName\t\"Lucida Sans\"");
            pw.println("\t\t]");
        }

        pw.println("\t]");
    }
    
    /**
     * Finishes the file.
     * 
     * @return true if successful
     */
    public boolean finish() {
        pw.println("]");
        pw.close();
        return !pw.checkError();
    }
    
        
    private static String wrap(String string) {
        StringBuilder wrapped = new StringBuilder(string.length());
        StringTokenizer tok = new StringTokenizer(string, "\n", true);
        while (tok.hasMoreTokens()) {
            String token = tok.nextToken();
            if (token.equals("\n")) {
                wrapped.append(token);
            } else {
                wrapped.append(wrapLine(token));                
            }
        }
        return wrapped.toString().replace('&', '+');
    }

    private static String wrapLine(String string) {
        int strLen = string.length();
        int restLength = strLen;

        StringBuilder wrapped = new StringBuilder(strLen);
        while (restLength > 0) {
            int nextLength = Math.min(PREFERRED_LINE_SIZE, restLength);
            restLength -= nextLength;

            if (restLength > 0) {
                char nextBeginChar = string.charAt(strLen - restLength);
                char endChar = string.charAt(strLen - restLength - 1);
                while (restLength > 1 && endChar != '.' && endChar != '/' && 
                        endChar != ' ' && nextBeginChar != '<'  && nextBeginChar != '(')
                {
                    restLength--;
                    nextLength++;
                    endChar = nextBeginChar;
                    nextBeginChar = string.charAt(strLen - restLength);
                }
            }
            
            if (restLength < MIN_LINE_SIZE) {
                nextLength += restLength;
                restLength = 0;
            }
            int beginIndex = strLen - restLength - nextLength;
            wrapped.append(string.substring(beginIndex, beginIndex + nextLength));
            if (restLength > 0) {
                wrapped.append("\n");
            }
        }
        
        return wrapped.toString();
    }

    
    

}
