/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.model.model;

import java.util.Iterator;

import org.eclipse.jdt.core.dom.IMethodBinding;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.intoJ.inferType.model.collection.DESet_Key;
import org.intoJ.inferType.model.collection.DESet_Signature;

/**
 * 
 * Abstract superclass representing declaration elements (declarations like
 * fields and methods in types) and invocations (field and method invocations on
 * declaration elements or other invocations)
 * 
 * This class is a wrapper class for IVariableBindings and IMethodBindings.
 * 
 * Please note that the standard equals method compares the keys of bindings,
 * whereas the method equalsSignature compares elements according to their
 * signature. For example, a method and its overriding method may be equal
 * according to signature, but not to key. See DESet_Key and DESet_Signature for
 * more information.
 * 
 * The declared type of the DE is the method return type for methods and the
 * variable type for variables. Note that this type may be an array type.
 * 
 * 
 * FAQ: 
 * 
 * Q: Why do we need a wrapper for IBinding? 
 * 
 * A: IVariableBindings representing parameters and local variables do not 
 * possess a DeclaringType.  However, we need that type to find the invocations 
 * on the elements.
 * 
 * @version: $Id: DeclarationElement.java,v 1.1 2007/09/09 19:40:25 kegel Exp $
 * @author Philip Mayer
 */
public abstract class DeclarationElement {

	private DESet_Signature localInvocations = null;
	private DESet_Key visitedDEs = null;
	private DESet_Key outAssignments = null;

	// For method declarations:
	private String returnType = null;

	/**
	 * Constants for getKind()
	 */

	public final static int PACKAGE = 0;
	public final static int CLASS = 1;
	public final static int INTERFACE = 2;
	/**
	 * The binding of a field is a IVariableBinding. It is declared as a FieldDeclaration.
	 */
	public final static int FIELD = 3;
	/**
	 * The binding of a method is an IMethodBinding. It is declared as a MethodDeclaration.
	 */
	public final static int METHOD = 4;
	/**
	 * The binding of a local variable (=method or initializer local, "for" loops) and a "catch clause variable" is a IVariableBinding.
	 * ->Method or initializer locals are declared as VariableDeclarationFragments inside of a VariableDeclarationStatement.
	 * ->"For" loops are declared as VariableDeclarationFragments too, but inside of a VariableDeclarationExpression.
	 * ->"Catch clause" variables are declared as SingleVariableDeclarations
	 * 
	 */
	public final static int LOCALVARIABLE = 5;
	/**
	 * The binding of a parameter is a IVariableBinding. It is declared as a SingleVariableDeclarations.
	 * Parameters are the only DEs to possess an index ID and a method parent.
	 */
	public final static int PARAMETER = 6;
	public final static int UNKNOWN = 7;

	public static final String[] KINDS = { "package", "class", "interface", "field", "method", "temporary", "parameter", "unknown" };

	/**
	 * Returns the simple name of this declaration element (method name or
	 * variable name).
	 * 
	 * @return String
	 */
	public abstract String getSimpleName();

	/**
	 * Returns the beautified name of this declaration element (method name and
	 * parameters or variable name).
	 * 
	 * @return String
	 */
	public abstract String getBeautifiedName();

	/**
	 * Returns the complete information on this DE 
	 * @return
	 */
	public abstract String getCompleteInformation();

	/**
	 * Returns the qualified name of this de's type (method return type or
	 * variable type)
	 * 
	 * @return String
	 */
	public abstract String getFullyQualifiedNameOfType();

	/**
	 * Type and beautified name of this declaration element.
	 * @return String
	 */
	public abstract String toString();

	/**
	 * Compares this declaration element to another using getKey() of the
	 * respective bindings.
	 * @param o Object
	 * @return boolean
	 */
	public abstract boolean equals(Object o);

	/**
	 * Hash Code generated via toString().
	 * @return int
	 */
	public abstract int hashCode();

	/**
	 * Compares Declaration Elements according to their signature. This differs
	 * from the equals method: Overridden methods, fields, and parameters are
	 * equal according to equalsSignature, but not according to equals.
	 * 
	 * @param d
	 * @return boolean
	 */
	public abstract boolean equalsSignature(DeclarationElement d);

	/**
	 * Returns one of these constant values (defined in DeclarationElement).
	 * 
	 * PACKAGE CLASS INTERFACE FIELD METHOD LOCALVARIABLE PARAM UNKNOWN
	 *  
	 * @return int
	 */
	public abstract int getKind();

	/**
	 * Returns the name of the kind of the declaration element.
	 *  
	 * @return String
	 */
	public String getKindAsString() {

		return KINDS[getKind()];
	}

	/**
	 * 
	 * Returns the declared type of this element. In case of methods, this is
	 * the return type; in case of variables, it is the type of the variable.
	 * 
	 * @return ITypeBinding
	 */
	public abstract ITypeBinding getDeclaredType();

	/**
	 * Returns the starting character number inside the source file of this
	 * declaration elements declaring class file.
	 * 
	 * 
	 * @return int
	 */
	public abstract int getStartPosition();

	/**
	 * Returns the modifiers (@see IBinding.getModifiers()).
	 * 
	 * @return int
	 */
	public abstract int getModifiers();

	/**
	 * Checks whether this declaration element has the given key. In case of
	 * Single Declaration Elements, the key is compared directly; in case of
	 * Composite Declaration Elements, this method returns true if one of the
	 * bindings contained within has this key.
	 * 
	 * @param key
	 * @return boolean
	 */
	public abstract boolean hasKey(String key);

	/**
	 * True if the declaration with this simple name exists once in this set.
	 * 
	 * @param name
	 * @return boolean
	 */
	public abstract boolean containsDeclarationOnce(String name);

	/**
	 * Sets the local invocations of the declaration element.
	 * 
	 * @param localInvocations the local invocations 
	 */
	public void setLocalInvocations(DESet_Key localInvocations) {

		this.localInvocations = new DESet_Signature(localInvocations);
	}

	/**
	 * Returns the local invocations of the declaration element.
	 * 
	 * @return local invocations
	 */
	public DESet_Signature getLocalInvocations() {

		return this.localInvocations;
	}

	/**
	 * Returns the transitive invocations of the declaration element. All of the Declaration
	 * Elements and their relationships are already present in the data structure, the concrete
	 * list, however, need to be created first by
	 * 
	 * 1. Adding all out assignment DEs of this declaration element
	 * 2. go through all the out assignments of those DEs and do 1.
	 * 
	 * Every DE MUST(!) already have its "localInvocation" list filled.
	 * 
	 * 
	 * @return transitive invocations (possibly null)
	 */
	public DESet_Signature getTransitiveInvocations() {

		DESet_Signature transitiveInvocations = new DESet_Signature();
		visitedDEs = new DESet_Key();
		determineTransitiveInvocations(this, transitiveInvocations);
		return transitiveInvocations;
	}

	/**
	 * This method recursively determines the transitive invocations for the given
	 * declaration element.
	 * 
	 * @param de the declaration element
	 * @param invocs the transitive invocations to be filled
	 */
	private void determineTransitiveInvocations(DeclarationElement de, DESet_Signature invocs) {

		if (visitedDEs.containsDeclaration(de))
			return;
		visitedDEs.addDeclaration(de);

		DESet_Signature li = de.getLocalInvocations();
		DESet_Key oa = de.getOutAssignments();

		invocs.addDeclarations(li);
		for (Iterator iter = oa.deIterator(); iter.hasNext();)
			determineTransitiveInvocations((DeclarationElement) iter.next(), invocs);
	}

	/**
	 * Sets the outAssignments of the declaration element.
	 * 
	 * @param transitiveOutAssignments DESet_Key
	 */
	public void setOutAssignments(DESet_Key transitiveOutAssignments) {

		this.outAssignments = transitiveOutAssignments;
	}

	/**
	 * Returns the outAssignments of the declaration element.
	 * 
	 * @return outAssignments (possibly null)
	 */
	public DESet_Key getOutAssignments() {

		return outAssignments;
	}

	/**
	 * Returns the transitive outAssignments of the declaration element.
	 * 
	 * @return transitive outAssignments (possibly null)
	 */
	public DESet_Key getTransitiveOutAssignments() {

		DESet_Key transitiveOutAssignments = new DESet_Key();
		visitedDEs = new DESet_Key();
		determineTransitiveOutAssignments(this, transitiveOutAssignments);
		return transitiveOutAssignments;
	}

	/**
	 * This method determines recursively the transitive outAssignments for the given
	 * declaration element.
	 * 
	 * @param de the declaration element
	 * @param outAssignments the transitive outAssignments to be filled
	 */
	private void determineTransitiveOutAssignments(DeclarationElement de, DESet_Key outAssignments) {

		if (visitedDEs.containsDeclaration(de))
			return;
		visitedDEs.addDeclaration(de);
		if(de.getOutAssignments() == null) return;
		outAssignments.addDeclarations(de.getOutAssignments());
		for (Iterator iter = de.getOutAssignments().deIterator(); iter.hasNext();)
			determineTransitiveOutAssignments((DeclarationElement) iter.next(), outAssignments);
	}

	/**
	 * Returns a String with the full qualified names of all declaring classes
	 * of the declaration element. 
	 * 
	 * Note that a composite declaration element has more than one declaring class.
	 * 
	 * @return full qualified names of all declaring classes
	 */
	public abstract String getDeclaringClassesToString();

	/**
	 * Sets the new declared type of the declaration element.
	 * The declared type of an declaration element is in case of methods
	 * the return type; in case of variables, the type of the variable.
	 * 
	 * The new declared type is needed for the refactoring.
	 * 
	 * @param newDeclaredType
	 */
	public abstract void setNewDeclaredType(TypeElement newDeclaredType);

	/**
	 * Returns the new declared type of the declaration element.
	 * The declared type of an declaration element is in case of methods
	 * the return type; in case of variables, the type of the variable.
	 * 
	 * The new declared type is needed for the refactoring.
	 * 
	 * @return TypeElement
	 */
	public abstract TypeElement getNewDeclaredType();

	/**
	 * If the declaration element is a parameter the declaring method binding
	 * is returned. Otherwise null will be returned. 
	 * 
	 * @return the declaring method binding of a paramter (otherwise null)
	 */
	public abstract IMethodBinding getMethodBinding();

	/**
	 * If this DE represents a method declaration, this method will set
	 * its simple parameter type names.
	 * 
	 * This is needed due to changes to inferred type methods after they are
	 * officially created. It is used on former method invocations now part
	 * of an inferred type declaration.
	 * 
	 * @return
	 */
	public abstract void setMethodParameterTypes(String[] params);

	/**
	 * If this DE represents a method declaration, this method will return
	 * its simple parameter type names.
	 * 
	 * This is needed due to changes to inferred type methods after they are
	 * officially created. It is used on former method invocations now part
	 * of an inferred type declaration.
	 * 
	 * @return
	 */
	public abstract String[] getMethodParameterTypes();

	/**
	 * If this DE represents a method declaration, this method will set the
	 * return type of the method.
	 * 
	 * This is needed due to changes to inferred type methods after they are
	 * officially created. It is used on former method invocations now part
	 * of an inferred type declaration.
	 * 
	 * @param newName
	 */
	public void setReturnType(String type) {

		this.returnType = type;
	}

	/**
	 * If this DE represents a method declaration, this method will return the
	 * return type of the method.
	 * 
	 * This is needed due to changes to inferred type methods after they are
	 * officially created. It is used on former method invocations now part
	 * of an inferred type declaration.
	 * 
	 * @return
	 */
	public String getReturnType() {

		if (returnType != null)
			return returnType;
		else
			return getFullyQualifiedNameOfType();
	}

	/**
	 * If this is a method or a variable, this method will return
	 * the dimensions of the return type/declared type, if any.
	 * @return
	 */
	public int getArrayDimensionsOfType() {

		return getDeclaredType().getDimensions();
	}

}
