/*
 * This class file belongs to the IntoJ project. See enclosed license file for
 * licensing issues.
 * 
 * For more information, please visit http://www.intoj.org/
 *  
 */
package org.intoJ.inferType.util;

import java.util.Collection;
import java.util.Iterator;

import org.eclipse.core.runtime.IStatus;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.JavaConventions;
import org.eclipse.jdt.core.dom.ITypeBinding;
import org.intoJ.inferType.exception.ValidationException;

/**
 * 
 * This class provides utility functions concerning strings.
 * 
 * @version: $Date: 2007/09/09 19:40:25 $ $Author: kegel $ $Revision: 1.1 $
 * @author Andreas Mei&szlig;ner
 */
public class Util_TypeNames {

	/**
	 * Convenience method for determining the type name.
	 * 
	 * @param fullyQualifiedTypeName
	 * @return String
	 */
	public static String getSimpleTypeName(String fullyQualifiedTypeName) {

		return fullyQualifiedTypeName.substring(fullyQualifiedTypeName.lastIndexOf('.') + 1);
	}

	/**
	 * Convenience method for determining the package name.
	 * 
	 * @param fullyQualifiedTypeName
	 * @return String
	 */
	public static String getPackageName(String fullyQualifiedTypeName) {

		if (fullyQualifiedTypeName.lastIndexOf('.') < 0)
			return "";
		return fullyQualifiedTypeName.substring(0, fullyQualifiedTypeName.lastIndexOf('.'));
	}

	/**
	 * This method validates a given type name and checks for length.
	 * 
	 * @param typeName
	 * @throws ValidationException
	 */
	public static void validateJavaTypeName(String typeName) throws ValidationException {

		if (typeName.length() == 0)
			throw new ValidationException("Please enter a type name!");

		String upperCaseTypeName = typeName.toUpperCase();
		IStatus val = JavaConventions.validateJavaTypeName(upperCaseTypeName);
		if (!val.isOK()) {
			throw new ValidationException("The type name '" + typeName + "' is not valid in " + "regard to the Java conventions.\n" + val.getMessage());
		}
	}

	/**
	 * This method validates a given type name and checks for according to the
	 * java speciication.
	 * 
	 * @param typeName
	 * @throws ValidationException
	 */
	public static void validateJavaTypeNameAccordingToSpec(String typeName) throws ValidationException {

		IStatus val = JavaConventions.validateJavaTypeName(typeName);
		if (!val.isOK()) {
			throw new ValidationException("The type name '" + typeName + "' is not valid in " + "regard to the Java conventions.\n" + val.getMessage());
		}
	}

	/**
	 * This method validates a given package name.
	 * 
	 * @param packageName
	 * @throws ValidationException
	 */
	public static void validatePackageName(String packageName) throws ValidationException {

		if (packageName.equals(""))
			return;

		IStatus val = JavaConventions.validatePackageName(packageName);
		if (!val.isOK()) {
			throw new ValidationException("The package name '" + packageName + "' is not valid in " + "regard to the Java conventions.\n" + val.getMessage());
		}
	}

	/**
	 * This method validates if a given package name is zero.
	 * 
	 * @param packageName
	 * @throws ValidationException
	 */
	public static void validatePackageNameLengthForZero(String packageName) throws ValidationException {

		if (packageName.length() == 0)
			throw new ValidationException("If you don't want to create the new type in the default package,\nplease enter a package name!");
	}

	/**
	 * This method converts an collection of IJavaElements to a String.
	 * 
	 * @param collection
	 *            of IJavaElements
	 * @return String
	 */
	public static String javaElementCollectionToString(Collection collection) {

		StringBuffer result = new StringBuffer();
		for (Iterator i = collection.iterator(); i.hasNext();) {
			IJavaElement current = (IJavaElement) i.next();
			result.append(current.getElementName() + "\n");
		}
		return result.toString();
	}

	/**
	 * Very important method when comparing ITypes and ITypeBindings. There are
	 * two differences in the way fully qualified names are treated in those two
	 * classes: 
	 * 
	 * Fist difference: ITypes separate inner and local classes via the '$' character
	 * from the top level class and packages, whereas ITypeBindings use the '.'
	 * character. 
	 * 
	 * The second difference: ITypes return a fully qualfied name for
	 * all kinds of types including local ones; ITypeBindings do not - to get a
	 * fully qualified name for an ITypeBinding (local type), use this method.
	 * 
	 * Note that to retrieve this fully qualified name, the declaring class of the node
	 * type must be found. This type MAY be null (see .getDeclaringClass()). It this is
	 * the case, resolving is not possible.
	 * 
	 * Note that the separation character is '.', too, as this is the case for
	 * ITypeBindings representing inner classes.
	 * 
	 * 
	 * @param local
	 * @return String
	 */
	public static String getQualifiedNameForAllTypesOfTypes(ITypeBinding local) {

		StringBuffer tmp = new StringBuffer();
		if (local.getQualifiedName().equals("")) {
			if (local.getDeclaringClass() != null) {
				String ret = getQualifiedNameForAllTypesOfTypes(local.getDeclaringClass());
				tmp.append(ret + "." + local.getName());
			} else
				tmp.append(local.getName());
		} else
			tmp.append(local.getQualifiedName());
		return tmp.toString();
	}

}